# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The library of intrinsic functions of the QuAIRKit. 
NOT ALLOWED to be exposed to users.
"""

import copy
import functools
import io
import math
import warnings
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

import IPython
import numpy as np
import PIL
import torch
from torch.nn.parameter import Parameter

from .base import get_dtype, get_float_dtype, get_seed
from .state import MixedState, PureState, State, to_state


def _format_total_dim(num_systems: int, system_dim: Union[List[int], int]) -> int:
    r"""Given number of systems with dimension, check the dimension of the total system
    
    Args:
        num_systems: total number of systems
        system_dim: dimension of each system
        
    Returns:
        dimension of the total system
    
    """
    if isinstance(system_dim, int):
        return system_dim ** num_systems 
    
    assert len(system_dim) == num_systems, \
        f"Dimension of each system should be equal to the number of systems: received {system_dim} and {num_systems}"
    return math.prod(system_dim)


def _format_system_dim(total_dim: int, system_dim: Union[List[int], int]) -> List[int]:
    r"""Given total dimension with dimension of each system, check the dimension of each system
    
    Args:
        total_dim: dimension of the total system
        system_dim: dimension of each system (could be an int)
        
    Returns:
        dimension of each 
    
    """
    if isinstance(system_dim, int):
        num_system = int(math.log(total_dim, system_dim))
        system_dim = [system_dim] * num_system
        assert math.prod(system_dim) == total_dim, \
            f"Total system dimension {total_dim} does not match the dimension of each system {len(system_dim)}"
    else:
        assert math.prod(system_dim) == total_dim, \
            f"Dimension of each system should be equal to the total dimension: received {system_dim} and {total_dim}"
    return system_dim


def _format_sequential_idx(system_idx: Union[List[List[int]], List[int], int, str],
                           num_systems: int, num_acted_system: int) -> List[List[int]]:
    r"""Formatting the system indices in a sequential way

    Args:
        system_idx: input system indices
        num_systems: total number of systems
        num_acted_system: the number of systems that one operation acts on

    Note:
        The shape of output system indices are formatted as [# of vertical gates, num_acted_system].
    """
    if not isinstance(system_idx, str):
        return system_idx

    if num_acted_system == 1:
        if system_idx == 'full':
            system_idx = list(range(num_systems))
        elif system_idx == 'even':
            system_idx = list(range(num_systems, 2))
        elif system_idx == 'odd':
            system_idx = list(range(1, num_systems, 2))

    elif system_idx == 'cycle':
        assert num_systems >= num_acted_system, \
            f"# of qubits should be >= # of acted qubits: received {num_systems} and {num_acted_system}"
        system_idx = [
            list(range(idx, idx + num_acted_system))
            for idx in range(num_systems - num_acted_system)
        ]
        system_idx.extend(
            list(range(idx, num_systems))
            + list(range(idx + num_acted_system - num_systems))
            for idx in range(num_systems - num_acted_system, num_systems)
        )
    
    elif system_idx == 'linear':
        assert num_systems >= num_acted_system, \
            f"# of system should be >= # of acted systems: received {num_systems} and {num_acted_system}"

        system_idx = [
            list(range(idx, idx + num_acted_system))
            for idx in range(num_systems - num_acted_system + 1)
        ]
    
    return system_idx


def _format_operator_idx(
        system_idx: Union[List[List[int]], List[int], int], num_acted_system: int
) -> List[List[int]]:
    r"""Formatting the system indices that operations acts on

    Args:
        system_idx: input system indices
        num_acted_system: the number of systems that one operation acts on

    Note:
        The shape of output system indices are formatted as [# of vertical gates, num_acted_system].
    """
    
    system_idx = np.array(system_idx).reshape([-1, num_acted_system])
    assert system_idx.shape == np.unique(system_idx, axis=1).shape, \
        f"Operators do not allow repeated indices: received {system_idx.tolist()}"
    return system_idx.tolist()


def _format_param_shape(system_idx: List[List[int]], num_acted_param: int, 
                        param_sharing: bool, batch_size: int = 1) -> List[int]:
    r"""Formatting the shape of parameters of param gates

    Args:
        system_idx: list of input system indices
        num_acted_param: the number of parameters required for a single operation
        param_sharing: whether all operations are shared by the same parameter set
        batch_size: size of gate batch, -1 for variable batch size

    Note:
        The input ``system_idx`` must be formatted by ``_format_system_idx`` first.
        The shape of parameters are formatted as [len(system_idx), batch_size, num_acted_param].

    """
    return [1 if param_sharing else len(system_idx), batch_size, num_acted_param]


def _theta_generation(net: torch.nn.Module, param: Union[torch.Tensor, float, List[float]], 
                      system_idx: List[List[int]], num_acted_param: int, param_sharing: bool) -> None:
    r""" determine net.theta, and create parameter if necessary

    Args:
        net: neural network instance
        param: input theta
        system_idx: list of input system indices
        num_acted_param: the number of parameters required for a single operation
        param_sharing: whether all operations are shared by the same parameter set

    Note:
        ``param`` will be transformed to a parameter:
        - if ``param`` is ``None``
        
        or ``param`` will be added to the parameter list:
        - if ``param`` is a ParamBase instance
        
        or ``param`` will keep unchanged:
        - if ``param`` is a Tensor but not a ParamBase
        - if ``param`` is an array of floats
        
        ``param`` will be shared by all operations:
        - if ``param_sharing`` is True
        - if ``param`` is a float scalar
    """
    float_dtype = _get_float_dtype(net.dtype)
    if param is None:
        expect_shape = _format_param_shape(system_idx, num_acted_param, param_sharing, batch_size=1)
        theta = torch.rand(expect_shape, dtype=float_dtype) * 2 * math.pi
        net.register_parameter('theta', Parameter(theta))
        return
    
    if isinstance(param, (int, float)):
        expect_shape = _format_param_shape(system_idx, num_acted_param, param_sharing, batch_size=1)
        net.theta = torch.ones(expect_shape, dtype=float_dtype) * param
        return
    
    if isinstance(param, Parameter):
        batch_size = int(param.numel() // (1 if param_sharing else len(system_idx)) // num_acted_param)
        expect_shape = _format_param_shape(system_idx, num_acted_param, param_sharing, batch_size=batch_size)
        
        assert list(param.shape) == expect_shape, \
            f"Shape assertion failed for input parameter: receive {list(param.shape)}, expect {expect_shape}"
        assert param.dtype == (torch.float32 if float_dtype == torch.float32 else torch.float64), \
            f"Dtype assertion failed for input parameter: receive {param.dtype}, expect {float_dtype}"
        net.register_parameter('theta', param)
    
    elif isinstance(param, torch.Tensor):
        expect_shape = _format_param_shape(system_idx, num_acted_param, param_sharing, -1)
        net.theta = param.to(net.device, dtype=float_dtype).reshape(expect_shape)
    
    else:  # when param is an Iterable
        expect_shape = _format_param_shape(system_idx, num_acted_param, param_sharing, -1)
        net.theta = torch.tensor(param, dtype=float_dtype).view(expect_shape)


def _get_float_dtype(complex_dtype: torch.dtype) -> torch.dtype:
    if complex_dtype == torch.complex64:
        float_dtype = torch.float32
    elif complex_dtype == torch.complex128:
        float_dtype = torch.float64
    else:
        raise ValueError(
            f"The dtype should be torch.complex64 or torch.complex128: received {complex_dtype}")
    return float_dtype


def _get_complex_dtype(float_dtype: torch.dtype) -> torch.dtype:
    if float_dtype == torch.float32:
        complex_dtype = torch.complex64
    elif float_dtype == torch.float64:
        complex_dtype = torch.complex128
    else:
        raise ValueError(
            f"The dtype should be torch.float32 or torch.float64: received {float_dtype}")
    return complex_dtype


_ArrayLike = Union[np.ndarray, torch.Tensor]
_StateLike = Union[np.ndarray, torch.Tensor, PureState, MixedState]
_ParamLike = Union[np.ndarray, torch.Tensor, Iterable[float]]
_SingleParamLike = Union[_ParamLike, float]
_General = Union[_ArrayLike, _StateLike, _SingleParamLike]


def _type_fetch(data: _General, 
                ndim: Optional[int]= None) -> Union[str, Tuple[str, List[int]]]:
    r"""Fetch the type of ``data``

    Args:
        data: the input data, and datatype of which should be either ``numpy.ndarray``,
            ``torch.Tensor``, ``quairkit.State``, ``Iterable[float]`` or ``float``
            where the last two types will be considered as "tensor".
        ndim: the number of dimensions to be removed, used for batched data

    Returns:
        When ndim is not none, the returned tuple contains the following variables:
        - a string of datatype of ``data``, can be either ``"numpy"``, ``"tensor"``, ``"state"``, or ``"other"``
        - the batch dimension
        When ndim is none, the returned variable is just the string.

    """
    if isinstance(data, np.ndarray):
        return "numpy" if ndim is None else ("numpy", list(data.shape[:-ndim]))

    if isinstance(data, torch.Tensor):
        return "tensor" if ndim is None else list(data.shape[:-ndim])
    
    if isinstance(data, State):
        return "state" if ndim is None else ("state", list(data.batch_dim))

    assert ndim is None, \
        ("Cannot obtain batch dimension for data type other than " +
         f"np.ndarray, torch.Tensor or quairkit.State: received {type(data)}")
    return "other"


def _density_to_vector(rho: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
    r""" transform a density matrix to a state vector

    Args:
        rho: a density matrix (pure state)

    Returns:
        a state vector

    Raises:
        ValueError: the output state may not be a pure state

    """
    type_str = _type_fetch(rho)
    rho = _type_transform(rho, "tensor")
    eigval, eigvec = torch.linalg.eigh(rho)

    max_eigval = torch.max(eigval).item()
    err = np.abs(max_eigval - 1)
    if err > 1e-6:
        raise ValueError(
            f"the output state may not be a pure state, maximum distance: {err}")

    state = eigvec[:, torch.argmax(eigval)]

    return state.detach().numpy() if type_str == "numpy" else state


def _type_transform(data: _General, output_type: str, 
                    system_dim: Optional[Union[List[int], int]] = None) -> _StateLike:
    r""" transform the datatype of ``input`` to ``output_type``

    Args:
        data: data to be transformed, can be and datatype of which should be either ``numpy.ndarray``,
            ``quairkit.State``, ``float``, ``Iterable[float]`` or ``torch.Tensor``
        output_type: datatype of the output data, type is either ``"numpy"``, ``"state"``, ``"tensor"``
        system_dim: dimension of the system, used for transforming to state. Defaults to qubit case.

    Returns:
        the output data with expected type

    Raises:
        ValueError: does not support transformation to type.

    """
    current_type = _type_fetch(data)
    
    if current_type == "other":
        current_type = "tensor"
        data = torch.tensor(data)

    if output_type == "other":
        output_type = "tensor"
    else:
        assert output_type in {"numpy", "tensor", "state"}, \
            f"does not support transformation from {current_type} to type {output_type}"

    if current_type == output_type:
        return data


    if current_type == "numpy":
        if output_type == "tensor":
            return torch.from_numpy(data)
        
        #TODO remove when all state manipulation functions are implemented
        if system_dim is None:
            shape = data.shape
            system_dim = shape[-2] if (len(shape) >= 2 and shape[-1] == 1) else shape[-1]
        else:
            system_dim = _format_system_dim(data.shape[-1], system_dim)

        return to_state(data, system_dim=system_dim, eps=None)

    if current_type == "tensor":
        if output_type == "numpy":
            return data.detach().cpu().resolve_conj().numpy()
        
        #TODO remove when all state manipulation functions are implemented
        if system_dim is None:
            shape = data.shape
            system_dim = shape[-2] if (len(shape) >= 2 and shape[-1] == 1) else shape[-1]
        else:
            system_dim = _format_system_dim(data.shape[-1], system_dim)

        return to_state(data, system_dim=system_dim, eps=None)

    if current_type == "state":
        if output_type == "numpy":
            return data.numpy()
        return data._data.clone()


def _cnot_idx_fetch(num_qubits: int, qubits_idx: List[Tuple[int, int]]) -> List[int]:
    r"""
    Compute the CNOT index obtained by applying the CNOT gate without using matrix multiplication.

    Args:
        num_qubits: The total number of qubits in the system.
        qubits_idx: A list of tuples, where each tuple contains the indices of the two qubits
                    involved in the CNOT gate.

    Returns:
        List: A list of integers representing the decimal values of all binary strings
                obtained by applying the CNOT gate.
    """
    assert len(np.shape(qubits_idx)) == 2, \
        "The CNOT system_idx should be list of tuple of integers, e.g., [[0, 1], [1, 2]]."
    binary_list = [bin(i)[2:].zfill(num_qubits) for i in range(2 ** num_qubits)]
    system_idx_length = len(qubits_idx)
    for item in range(len(binary_list)):
        for bin_idx in range(system_idx_length):
            id1 = qubits_idx[system_idx_length - bin_idx - 1][0]
            id2 = qubits_idx[system_idx_length - bin_idx - 1][1]
            if binary_list[item][id1] == "1":
                if binary_list[item][id2] == '0':
                    binary_list[item] = binary_list[item][:id2] + '1' + binary_list[item][id2 + 1:]
                else:
                    binary_list[item] = binary_list[item][:id2] + '0' + binary_list[item][id2 + 1:]

    decimal_list = [int(binary, 2) for binary in binary_list]
    return decimal_list


def _assert_gradient(func: Union[Callable[[torch.Tensor], torch.Tensor], str], param: torch.Tensor, 
                     atol: float = 1e-2, rtol: float = 1e-2, eps: float = 1e-2,
                     feedback: bool = False, message: str = None) -> None:
    r"""
    Check the correctness of the gradient computation using finite difference method.

    Args:
        function_name: function to be tested.
        param: variables that perform finite differences.
        atol: absolute tolerance. Defaults to 1e-3.
        rtol: relative tolerance. Defaults to 1e-2.
        eps: perturbation for finite differences. Defaults to 1e-2.
        feedback: whether to print the result for confirmation. Defaults to False.
        message: optional message to be printed if the gradient check fails. Defaults to None.
    
    """
    seed, dtype = get_seed(), get_float_dtype()
    assert dtype == torch.float64, \
        "The gradient check only supports double precision, which can be set by `quairkit.set_dtype('complex128')`."
    param = param.to(dtype).clone().requires_grad_(True)
    
    try:
        result = torch.autograd.gradcheck(func, param, eps=eps, atol=atol, 
                                          rtol=rtol, raise_exception=True)
    except Exception as e:
        result = False
        warnings.warn("The following error occurred when trying to compute the gradient\n" 
                      + str(e), RuntimeWarning)
    
    assert result, (
        "The function is not accurate enough for precision \n"
        f"Setting: seed {seed}, precision {atol}, message {message} "
    )
        
    # Print the result for confirmation
    if feedback:
        print(f"Gradient test with seed {seed} passed",
              f"under abs tolerance {atol} and rel tolerance {rtol}.")


def _is_sample_linear(
    func: Callable[[Union[torch.Tensor, np.ndarray]], Union[torch.Tensor, np.ndarray]],
    info: Union[List[int], Callable[[], torch.Tensor], Callable[[], np.ndarray]],
    input_dtype: torch.dtype = None,
) -> Tuple[Callable[[torch.Tensor], torch.Tensor], Callable[[], torch.Tensor], torch.dtype, str]:
    r"""Check whether the inputs in `is_linear`, `create_matrix`, `create_choi_repr` are legal

    Args:
        func: A callable function to be tested.
        info: Information of the sample function, either the shape of sample data or the function itself.
        input_dtype: The data type of the input tensor. Defaults to None.

    Returns:
        A tuple containing:
        - A function to be tested, which ensures the input and output are torch.Tensors.
        - A function that generates input samples as torch.Tensors.
        - The data type of the input tensor.
        - A string representing the original type of the input function's output.

    Raises:
        TypeError: The input arguments are not of expected types.
    """
    assert isinstance(
        func, Callable
    ), f"The input 'func' must be callable, received {type(func)}"

    # Process the 'info' argument to create a tensor generator
    generator = info

    if input_dtype is None:
        input_dtype = get_dtype()
    elif not isinstance(input_dtype, torch.dtype):
        raise TypeError(f"Input must be of type torch.dtype or None. Received type: {type(input_dtype)}")

    if isinstance(info, list) and all((isinstance(x, int) and x > 0) for x in info):
        # If 'info' is a list of integers, create a torch.tensor generator based on the shape of input data and the input dtype
        generator = lambda: torch.randn(info, dtype=input_dtype)
        sample = generator()
        func_in = func

    elif isinstance(info, Callable):
        sample = info()
        # Handling different types of generated inputs

        if isinstance(sample, np.ndarray):
            generator = lambda: torch.from_numpy(info())
            func_in = lambda input_data: func(torch.from_numpy(input_data))
        elif isinstance(sample, torch.Tensor):
            generator = lambda: info()
            func_in = func
        else:
            raise TypeError(
                f"The output of info must be a torch.Tensor or numpy.ndarray: received {type(sample)}"
            )

    else:
        raise TypeError(
            "the info entry should either the shape of input data, or a Callable functions: "
            + f"received info {info} with type {type(info)}"
        )

    # guarantee the output type is torch.Tensor
    output = func(sample)

    func_tensor = func_in
    if isinstance(output, np.ndarray):
        func_tensor = lambda mat: torch.from_numpy(func_in(mat))
    elif not isinstance(output, torch.Tensor):
        raise TypeError(
            f"The input function either return torch.Tensor or np.ndarray: received type {type(output)}"
        )

    return func_tensor, generator, input_dtype, _type_fetch(output)


def _alias(aliases: Dict[str, str]) -> Callable:
    r"""alias decorator for function arguments
    
    Args:
        aliases: a dictionary of alias names. The key is the original name, and the value is the alias name.
    
    Returns:
        A decorator that replaces the alias name with the original name.
    
    Note:
        See stack overflow #29374425
    
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            for name, alias in aliases.items():
                if name not in kwargs and alias in kwargs:
                    kwargs[name] = copy.deepcopy(kwargs[alias])
                    del kwargs[alias]
            return func(*args, **kwargs)
        return wrapper
    return decorator


def _digit_to_int(digits: str, base: Union[int, List[int]]) -> int:
    r"""Convert a string of digits in a given base to an integer
    
    Args:
        digits: a string of digits to be converted
        base: the base of the digits, or a list of bases for each digit
    """
    if isinstance(base, int):
        return int(digits, base=base)
    
    number = 0
    multiplier = 1
    for digit, base in zip(reversed(digits), reversed(base)):
        if not digit.isdigit() or int(digit) >= base:
            raise ValueError(
                f"Digit '{digit}' is invalid for base {base}.")
        
        number += int(digit) * multiplier
        multiplier *= base   
    return number


def _int_to_digit(number: int, base: Union[int, List[int]]) -> str:
    r"""Convert an integer to a string of digits in a given base
    
    Args:
        number: the integer to be converted
        base: the base to convert the integer, or a list of bases for each digit
    """
    if number < 0:
        return f'-{_int_to_digit(-number, base)}'

    if isinstance(base, int):
        return np.base_repr(number, base=base)

    digits = []
    for base in reversed(base):
        digits.append(str(number % base))
        number //= base
    return ''.join(reversed(digits))


def _replace_indices(l: Iterable[int], 
                     m: Union[List[int], Dict[int, int]]) -> Iterable[int]:
    r"""Replace the indices in l with the corresponding elements in m
    
    Args:
        l: a nested list to be replaced, where each element is an index
        m: a list or a dictionary of replacement
        
    Returns:
        a mapped list
    """
    def replace_recursive(element):
        if isinstance(element, Iterable):
            return [replace_recursive(sub_element) for sub_element in element]
        else:
            return m[element]

    return replace_recursive(l)


def __is_jupyter() -> bool:
    r"""Check whether the current environment is Jupyter notebook
    
    Returns:
        True if the current environment is Jupyter notebook
    """
    try:
        from IPython import get_ipython
        return get_ipython().__class__.__name__ == 'ZMQInteractiveShell'
    except Exception:
        return False

def _display_png(img: IPython.display.Image) -> None:
    r"""Display input picture that is stored using IPython
    
    Args:
        img: the image stored in png to be displayed
    """
    if __is_jupyter():
        IPython.display.display(img)
    else:
        pil_img = PIL.Image.open(io.BytesIO(img.data))
        pil_img.show()



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the classes for multi-qubit gates.
"""

from typing import Iterable, List, Optional, Tuple, Union

import matplotlib
import torch

from ...core.utils.matrix import (_cnot, _cp, _crx, _cry, _crz, _cswap, _cu,
                                  _cy, _cz, _ms, _rxx, _ryy, _rzz, _swap,
                                  _toffoli, _universal2, _universal3)
from .base import Gate, ParamGate
from .visual import (_cnot_display, _crx_like_display, _cswap_display,
                     _cx_like_display, _oracle_like_display, _rxx_like_display,
                     _swap_display, _tofolli_display)


class CNOT(Gate):
    r"""A collection of CNOT gates.

    For a 2-qubit quantum circuit, when ``qubits_idx`` is ``[0, 1]``, the matrix form of such a gate is:

    .. math::

        \begin{align}
            CNOT &=|0\rangle \langle 0|\otimes I + |1 \rangle \langle 1|\otimes X\\
            &=
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 1 \\
                0 & 0 & 1 & 0
            \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
    
    """

    __matrix = _cnot(torch.complex128)

    def __init__(
        self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
    ):
        gate_info = {
            "name": "ctrl-x",
            "tex": r'\targ{}',
            "api": "cnot",
            "num_ctrl_system": 1,
            "label": '1',
            'plot_width': 0.2,
        }
        super().__init__(
            None, qubits_idx, acted_system_dim=[2, 2], check_legality=False, gate_info=gate_info)
    
    @property
    def matrix(self) -> torch.Tensor:
        return CNOT.__matrix.to(self.device, dtype=self.dtype)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _cnot_display(self, ax, x, )

CX = CNOT


class CY(Gate):
    r"""A collection of controlled Y gates.

    For a 2-qubit quantum circuit, when ``qubits_idx`` is ``[0, 1]``, the matrix form of such a gate is:

    .. math::

        \begin{align}
            CY &=|0\rangle \langle 0|\otimes I + |1 \rangle \langle 1|\otimes Y\\
            &=
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & -1j \\
                0 & 0 & 1j & 0
            \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
    
    """

    __matrix = _cy(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
    ):
        gate_info = {
            "name": "ctrl-y",
            "tex": r'Y',
            "api": "cy",
            "num_ctrl_system": 1,
            "label": '1',
            'plot_width': 0.4,
        }
        super().__init__(
            None, qubits_idx, acted_system_dim=[2, 2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return CY.__matrix.to(self.device, dtype=self.dtype)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _cx_like_display(self, ax, x, )


class CZ(Gate):
    r"""A collection of controlled Z gates.

    For a 2-qubit quantum circuit, when ``qubits_idx`` is ``[0, 1]``, the matrix form of such a gate is:

    .. math::

        \begin{align}
            CZ &=|0\rangle \langle 0|\otimes I + |1 \rangle \langle 1|\otimes Z\\
            &=
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                0 & 0 & 0 & -1
            \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
    
    """

    __matrix = _cz(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
    ):
        gate_info = {
            "name": "ctrl-z",
            "tex": r'Z',
            "api": "cz",
            "num_ctrl_system": 1,
            "label": '1',
            'plot_width': 0.4,
        }
        super().__init__(
            None, qubits_idx, acted_system_dim=[2, 2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return CZ.__matrix.to(self.device, dtype=self.dtype)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _cx_like_display(self, ax, x)


class SWAP(Gate):
    r"""A collection of SWAP gates.

    The matrix form of such a gate is:

    .. math::

        \begin{align}
            SWAP =
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 1
            \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
    
    """

    __matrix = _swap(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
    ):
        gate_info = {
            "name": "qubit-swap",
            "api": "swap",
            "permute": [1, 0],
            'plot_width': 0.2,
        }
        super().__init__(
            None, qubits_idx, acted_system_dim=[2, 2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return SWAP.__matrix.to(self.device, dtype=self.dtype)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _swap_display(self, ax, x, )


class CP(ParamGate):
    r"""A collection of controlled P gates.

    For a 2-qubit quantum circuit, when ``qubits_idx`` is ``[0, 1]``, the matrix form of such a gate is:

    .. math::

        \begin{bmatrix}
            1 & 0 & 0 & 0\\
            0 & 1 & 0 & 0\\
            0 & 0 & 1 & 0\\
            0 & 0 & 0 & e^{i\theta}
        \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    
    """

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, str]] = None,
            param: Optional[Union[torch.Tensor, float]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "ctrl-p",
            "tex": r'P',
            "api": "cp",
            "num_ctrl_system": 1,
            "label": '1',
            'plot_width': 0.9,
        }

        super().__init__(
            _cp, param, 1, param_sharing, qubits_idx, [2, 2], check_legality=False, gate_info=gate_info, support_batch=True)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _crx_like_display(self, ax, x)


class CRX(ParamGate):
    r"""A collection of controlled rotation gates about the x-axis.

    For a 2-qubit quantum circuit, when ``qubits_idx`` is ``[0, 1]``, the matrix form of such a gate is:

    .. math::

        \begin{align}
            CRx &=|0\rangle \langle 0|\otimes I + |1 \rangle \langle 1|\otimes Rx\\
            &=
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} \\
                0 & 0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2}
            \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    
    """

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, str]] = None,
            param: Optional[Union[torch.Tensor, float]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "ctrl-rx",
            "tex": r'R_{x}',
            "api": "crx",
            "num_ctrl_system": 1,
            "label": '1',
            "param_sharing": param_sharing,
            'plot_width': 0.9,
        }

        super().__init__(
            _crx, param, 1, param_sharing, qubits_idx, [2, 2], check_legality=False, gate_info=gate_info, support_batch=True)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _crx_like_display(self, ax, x)


class CRY(ParamGate):
    r"""A collection of controlled rotation gates about the y-axis.

    For a 2-qubit quantum circuit, when ``qubits_idx`` is ``[0, 1]``, the matrix form of such a gate is:

    .. math::

        \begin{align}
            CRy &=|0\rangle \langle 0|\otimes I + |1 \rangle \langle 1|\otimes Ry\\
            &=
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & \cos\frac{\theta}{2} & -\sin\frac{\theta}{2} \\
                0 & 0 & \sin\frac{\theta}{2} & \cos\frac{\theta}{2}
            \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    
    """

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, str]] = None,
            param: Optional[Union[torch.Tensor, float]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "ctrl-ry",
            "tex": r'R_{y}',
            "api": "cry",
            "num_ctrl_system": 1,
            "label": '1',
            "param_sharing": param_sharing,
            'plot_width': 0.9,
        }

        super().__init__(
            _cry, param, 1, param_sharing, qubits_idx, [2, 2], check_legality=False, gate_info=gate_info, support_batch=True)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _crx_like_display(self, ax, x, )


class CRZ(ParamGate):
    r"""A collection of controlled rotation gates about the z-axis.

    For a 2-qubit quantum circuit, when ``qubits_idx`` is ``[0, 1]``, the matrix form of such a gate is:

    .. math::

        \begin{align}
            CRz &=|0\rangle \langle 0|\otimes I + |1 \rangle \langle 1|\otimes Rz\\
            &=
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & e^{-i\frac{\theta}{2}} & 0 \\
                0 & 0 & 0 & e^{i\frac{\theta}{2}}
            \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    
    """

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, str]] = None,
            param: Optional[Union[torch.Tensor, float]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "ctrl-rz",
            "tex": r'R_{z}',
            "api": "crz",
            "num_ctrl_system": 1,
            "label": '1',
            "param_sharing": param_sharing,
            'plot_width': 0.9,
        }

        super().__init__(
            _crz, param, 1, param_sharing, qubits_idx, [2, 2], check_legality=False, gate_info=gate_info, support_batch=True)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _crx_like_display(self, ax, x)


class CU(ParamGate):
    r"""A collection of controlled single-qubit rotation gates.

    For a 2-qubit quantum circuit, when ``qubits_idx`` is ``[0, 1]``, the matrix form of such a gate is:

    .. math::

        \begin{align}
            CU
            &=
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & e^{i\gamma}\cos\frac\theta2 &-e^{i(\lambda+\gamma)}\sin\frac\theta2 \\
                0 & 0 & e^{i(\phi+\gamma)}\sin\frac\theta2&e^{i(\phi+\lambda+\gamma)}\cos\frac\theta2
            \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    
    """

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, str]] = None,
            param: Optional[Union[torch.Tensor, float]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "ctrl-u4",
            "tex": r'U',
            "api": "cu",
            "num_ctrl_system": 1,
            "label": '1',
            "param_sharing": param_sharing,
            'plot_width': 1.65,
        }
        super().__init__(
            _cu, param, 4, param_sharing, qubits_idx, [2, 2], check_legality=False, gate_info=gate_info, support_batch=True)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _crx_like_display(self, ax, x)


class RXX(ParamGate):
    r"""A collection of RXX gates.

    The matrix form of such a gate is:

    .. math::

        \begin{align}
            RXX(\theta) =
                \begin{bmatrix}
                    \cos\frac{\theta}{2} & 0 & 0 & -i\sin\frac{\theta}{2} \\
                    0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} & 0 \\
                    0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2} & 0 \\
                    -i\sin\frac{\theta}{2} & 0 & 0 & \cos\frac{\theta}{2}
                \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    
    """

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, str]] = None,
            param: Optional[Union[torch.Tensor, float]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "rxx",
            "tex": r'R_{xx}',
            "api": "rxx",
            "param_sharing": param_sharing,
            'plot_width': 1.0,
        }
        super().__init__(
            _rxx, param, 1, param_sharing, qubits_idx, [2, 2], check_legality=False, gate_info=gate_info, support_batch=True)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _rxx_like_display(self, ax, x)


class RYY(ParamGate):
    r"""A collection of RYY gates.

    The matrix form of such a gate is:

    .. math::

        \begin{align}
            RYY(\theta) =
                \begin{bmatrix}
                    \cos\frac{\theta}{2} & 0 & 0 & i\sin\frac{\theta}{2} \\
                    0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} & 0 \\
                    0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2} & 0 \\
                    i\sin\frac{\theta}{2} & 0 & 0 & cos\frac{\theta}{2}
                \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    
    """

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, str]] = None,
            param: Optional[Union[torch.Tensor, float]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "ryy",
            "tex": r'R_{yy}',
            "api": "ryy",
            "param_sharing": param_sharing,
            'plot_width': 1.0,
        }

        super().__init__(
            _ryy, param, 1, param_sharing, qubits_idx, [2, 2], check_legality=False, gate_info=gate_info, support_batch=True)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _rxx_like_display(self, ax, x)


class RZZ(ParamGate):
    r"""A collection of RZZ gates.

    The matrix form of such a gate is:

    .. math::

        \begin{align}
            RZZ(\theta) =
                \begin{bmatrix}
                    e^{-i\frac{\theta}{2}} & 0 & 0 & 0 \\
                    0 & e^{i\frac{\theta}{2}} & 0 & 0 \\
                    0 & 0 & e^{i\frac{\theta}{2}} & 0 \\
                    0 & 0 & 0 & e^{-i\frac{\theta}{2}}
                \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    
    """

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, str]] = None,
            param: Optional[Union[torch.Tensor, float]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "rzz",
            "tex": r'R_{zz}',
            "api": "rzz",
            "param_sharing": param_sharing,
            'plot_width': 1.0,
        }
        super().__init__(
            _rzz, param, 1, param_sharing, qubits_idx, [2, 2], check_legality=False, gate_info=gate_info, support_batch=True)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _rxx_like_display(self, ax, x)


class MS(Gate):
    r"""A collection of Mølmer-Sørensen (MS) gates for trapped ion devices.

    The matrix form of such a gate is:

    .. math::

        \begin{align}
            MS = RXX(-\frac{\pi}{2}) = \frac{1}{\sqrt{2}}
                \begin{bmatrix}
                    1 & 0 & 0 & i \\
                    0 & 1 & i & 0 \\
                    0 & i & 1 & 0 \\
                    i & 0 & 0 & 1
                \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first two qubits.
    
    """

    __matrix = _ms(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None
    ):
        gate_info = {
            "name": "ms",
            "tex": r'\text{MS}',
            "api": "ms",
            'plot_width': 0.6,
        }
        super().__init__(
            None, qubits_idx, acted_system_dim=[2, 2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return MS.__matrix.to(self.device, dtype=self.dtype)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _oracle_like_display(self, ax, x)


class CSWAP(Gate):
    r"""A collection of CSWAP (Fredkin) gates.

    The matrix form of such a gate is:

    .. math::

        \begin{align}
            CSWAP =
            \begin{bmatrix}
                1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 \\
                0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & 0 & 1
            \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first three qubits.
    
    """
    __matrix = _cswap(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None
    ):
        gate_info = {
            "name": "ctrl-swap",
            "api": "cswap",
            "num_ctrl_system": 1,
            "label": '1',
            "permute": [1, 0],
            'plot_width': 0.2,
        }
        super().__init__(
            None, qubits_idx, acted_system_dim=[2, 2, 2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return CSWAP.__matrix.to(self.device, dtype=self.dtype)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _cswap_display(self, ax, x)


class CCX(Gate):
    r"""A collection of CCX (Toffoli) gates.

    The matrix form of such a gate is:

    .. math::

        \begin{align}
            \begin{bmatrix}
                1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 \\
                0 & 0 & 0 & 0 & 0 & 0 & 1 & 0
            \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first three qubits.
    
    """
    __matrix = _toffoli(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None
    ):
        gate_info = {
            "name": "toffoli",
            "tex": r'\targ{}',
            "api": "ccx",
            "num_ctrl_system": 2,
            "label": '11',
            'plot_width': 0.2,
        }

        super().__init__(
            None, qubits_idx, acted_system_dim=[2, 2, 2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return CCX.__matrix.to(self.device, dtype=self.dtype)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _tofolli_display(self, ax, x, )


Toffoli = CCX



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Linear algebra functions in QuAIRKit.
"""

import math
from collections import OrderedDict, defaultdict
from functools import reduce
from itertools import product
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import scipy
import torch


def _abs_norm(mat: torch.Tensor) -> float: 
    norms = torch.norm(torch.abs(mat), dim=(-2, -1))
    return norms.item() if mat.ndim == 2 else norms.tolist()


def _p_norm_herm(mat: torch.Tensor, p: torch.Tensor) -> torch.Tensor:
    eigval = torch.linalg.eigvalsh(mat)
    norm = torch.abs(eigval).pow(p).sum(dim = len(list(mat.shape[:-2]))).pow(1 / p)
    return norm.view(mat.shape[:-2])


def _p_norm(mat: torch.Tensor, p: torch.Tensor) -> torch.Tensor:
    s = torch.linalg.svdvals(mat)
    norm = s.pow(p).sum(dim = len(list(mat.shape[:-2]))).pow(1 / p) 
    return norm.view(mat.shape[:-2])


def _dagger(mat: torch.Tensor) -> torch.Tensor: 
    return mat.mH.contiguous()


def _block_enc_herm(   
    mat: torch.Tensor, num_block_qubits: Optional[int] = 1
) -> torch.Tensor:  
    H = mat.detach().numpy()
    complex_dtype = mat.dtype

    num_qubits = int(math.log2(mat.shape[0]))
    H_complement = scipy.linalg.sqrtm(np.eye(2**num_qubits) - H @ H)
    block_enc = np.block([[H, 1j * H_complement], [1j * H_complement, H]])
    block_enc = torch.from_numpy(block_enc.astype("complex128")).to(complex_dtype)

    if num_block_qubits > 1:
        block_enc = _direct_sum(
            block_enc,
            torch.eye(2 ** (num_block_qubits + num_qubits) - 2 ** (num_qubits + 1)).to(
                complex_dtype
            ),
        )

    return block_enc



def _direct_sum(A: torch.Tensor, B: torch.Tensor) -> torch.Tensor: 
    
    batch_dim = list(A.shape[:-2])
    batch_size = [int(np.prod(batch_dim))] 
    
    A = A.reshape(batch_size + list(A.shape[-2:]))
    B = B.reshape(batch_size + list(B.shape[-2:]))
    
    A_rows, A_cols = A.shape[-2], A.shape[-1]
    B_rows, B_cols = B.shape[-2], B.shape[-1]

    batch_size = int(np.prod(batch_dim))
    # Create zero matrices with appropriate shapes
    zero_AB = torch.zeros((batch_size, A_rows, B_cols), dtype=A.dtype)
    zero_BA = torch.zeros((batch_size, B_rows, A_cols), dtype=B.dtype)

    # Stack matrices to form the direct sum
    mat_upper = torch.cat((A, zero_AB), dim=-1)
    mat_lower = torch.cat((zero_BA, B), dim=-1)
    mat = torch.cat((mat_upper, mat_lower), dim=-2)

    return mat.squeeze()


def _herm_transform( 
    fcn: Callable[[float], float],
    mat: torch.Tensor,
    ignore_zero: Optional[bool] = False,
) -> torch.Tensor:
    eigval, eigvec = torch.linalg.eigh(mat)
    eigval = eigval.tolist()
    eigvec = eigvec.T

    mat = torch.zeros(mat.shape).to(mat.dtype)
    for i in range(len(eigval)):
        vec = eigvec[i].reshape([mat.shape[0], 1])

        if np.abs(eigval[i]) < 1e-5 and ignore_zero:
            continue
        mat += (fcn(eigval[i]) + 0j) * vec @ torch.conj(vec.T)

    return mat


def _subsystem_decomposition(
    mat: torch.Tensor,
    first_basis: List[torch.Tensor],
    second_basis: List[torch.Tensor],
    inner_prod: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
) -> torch.Tensor:
    # Check if the input has a batch dimension
    if mat.ndim == 3:
        batch_size, d1, d2 = mat.shape
    else:
        d1, d2 = mat.shape
        batch_size = 1
        mat = mat.unsqueeze(0)  # Add a batch dimension

    assert (
        (d1, d2) == torch.kron(first_basis[0], second_basis[0]).shape
    ), f"The shape does not agree: received {mat.shape, first_basis[0].shape, second_basis[0].shape}"

    first_dim, second_dim = len(first_basis), len(second_basis)

    # Create Kronecker products for all pairs in the basis sets
    kron_products = torch.stack([
        torch.kron(first_basis[i], second_basis[j])
        for i, j in product(range(first_dim), range(second_dim))
    ])

    # Expand dimensions to match the batch size
    kron_products = kron_products.unsqueeze(0).expand(batch_size, -1, -1, -1)

    # Reshape for batch matrix multiplication
    mat = mat.unsqueeze(1).expand(-1, first_dim * second_dim, -1, -1)
    kron_products = kron_products.reshape(batch_size * first_dim * second_dim, d1, d2)
    mat = mat.reshape(batch_size * first_dim * second_dim, d1, d2)

    # Compute inner product in a batched way
    coefs = inner_prod(kron_products, mat).unsqueeze(0) 

    # Reshape the result to [batch_size, first_dim, second_dim]
    coefs = coefs.view(batch_size, first_dim, second_dim)

    return coefs.squeeze(0) if batch_size == 1 else coefs


def _perm_to_left(mat: torch.Tensor, perm_system_idx: List[int], system_dim: List[int]) -> torch.Tensor:
    r"""Permute the given systems of input matrix to the left.
    
    Args:
        mat: input matrix.
        perm_system_idx: the indices of the systems to be permuted.
        system_dim: the dimensions of all systems.
    
    """
    num_systems = len(system_dim)
    if perm_system_idx == list(range(num_systems)):
        return mat
    target_idx = perm_system_idx + [x for x in list(range(num_systems)) if x not in perm_system_idx]
    return _permute_systems(mat, target_idx, system_dim)


def _trace_1(mat: torch.Tensor, dim1: int) -> torch.Tensor:
    r"""Tracing out the first system of the matrix
    
    Args:
        dim1: the dimension of the first system.
    
    """
    batch_dims, dim2 = list(mat.shape[:-2]), int(mat.shape[-1] // dim1)
    mat = mat.view(-1, dim1, dim2, dim1, dim2)
    return _trace(mat, axis1=-2, axis2=-4).reshape(batch_dims + [dim2, dim2])


def _ptrace_1(xy: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
    r"""Tracing out the first system of the product unit vector xy
    
    Args:
        x: the unit vector of the first system.
    
    """
    dim_x = x.shape[-1]
    batch_dims, dim_y = list(xy.shape[:-1]), int(xy.shape[-1] // dim_x)
    
    xy, x = xy.view([-1, dim_x, dim_y]), x.view([-1, dim_x, 1])
    y = (xy * x.conj()).sum(dim=-2)
    return y.view(batch_dims + [dim_y])


def _transpose_1(mat: torch.Tensor, dim1: int) -> torch.Tensor:
    r"""Transpose the first system of the matrix
    
    Args:
        dim1: the dimension of the first system.
    
    """
    total_dim = mat.shape[-1]
    batch_dims, dim2 = list(mat.shape[:-2]), int(total_dim // dim1)
    mat = mat.view(-1, dim1, dim2, dim1, dim2)
    return torch.permute(mat, [0, 3, 2, 1, 4]).reshape(batch_dims + [total_dim, total_dim])


def _partial_trace(
    state: torch.Tensor, trace_idx: List[int], system_dim: List[int]
) -> torch.Tensor:
    state = _perm_to_left(state, trace_idx, system_dim)
    
    traced_dim = int(np.prod([system_dim[i] for i in trace_idx]))
    return _trace_1(state, traced_dim)


def _partial_trace_discontiguous(  
    rho: torch.Tensor, preserve_qubits: List[int]
) -> torch.Tensor:
    if preserve_qubits is None:
        return rho

    n = int(math.log2(rho.shape[-1]))
    system_dim, traced_qubits = [2] * n, [i for i in range(n) if i not in preserve_qubits]
    return _partial_trace(rho, traced_qubits, system_dim)


def _density_to_vector(rho: torch.Tensor) -> torch.Tensor: 

    # Handle a batch of density matrices
    batch_dim = list(rho.shape[:-2])
    batch_size = [int(np.prod(batch_dim))]

    rho_size = [rho.shape[-1]]
    eigval, eigvec = torch.linalg.eigh(rho)
    eigvec = eigvec.reshape(batch_size + 2*rho_size)
    eigval = eigval.reshape(batch_size + rho_size)

    max_eigval_values, max_eigval_indices = torch.max(eigval, dim=1)

    eig_test = torch.abs(max_eigval_values - 1) > 1e-6
    if torch.any(eig_test):
        raise ValueError(
            f"The output state may not be a pure state, maximum eigenvalue distance: {torch.abs(max_eigval_values - 1)}"
        )

    return eigvec[torch.arange(len(max_eigval_indices)), : ,max_eigval_indices].squeeze().unsqueeze(-1)


def _trace(mat: torch.Tensor, axis1: int = -2, axis2: int =- 1) -> torch.Tensor:
    dia_elements = torch.diagonal(mat, offset=0 ,dim1=axis1, dim2=axis2)
    return torch.sum(dia_elements, dim=-1)


def _kron(matrix_A: torch.Tensor, matrix_B: torch.Tensor) -> torch.Tensor:
    r"""(batched) Kronecker product
    
    Args:
        matrix_A: input (batched) matrix
        matrix_B: input (batched) matrix
    
    Returns:
        The Kronecker product of the two (batched) matrices
    
    Note:
        See https://discuss.pytorch.org/t/kronecker-product/3919/11
    """
    mat_dim = torch.Size([matrix_A.shape[-2] * matrix_B.shape[-2], matrix_A.shape[-1] * matrix_B.shape[-1]])
    output = matrix_A.unsqueeze(-1).unsqueeze(-3) * matrix_B.unsqueeze(-2).unsqueeze(-4)
    batch_dim = output.shape[:-4]
    return output.reshape(batch_dim + mat_dim)


def _nkron( 
    matrix_1st: torch.Tensor, *args: torch.Tensor
) -> torch.Tensor:
    return reduce(_kron, args, matrix_1st)


def _partial_transpose(state: torch.Tensor, transpose_idx: List[int], system_dim: List[int]) -> torch.Tensor: 
    state = _perm_to_left(state, transpose_idx, system_dim)
    transpose_dim = int(np.prod([system_dim[i] for i in transpose_idx]))
    state = _transpose_1(state, transpose_dim)

    original_seq = list(range(len(system_dim)))
    current_seq = transpose_idx + [i for i in original_seq if i not in transpose_idx]
    current_system_dim = [system_dim[x] for x in current_seq]
    map_original = _perm_of_list(current_seq, original_seq)
    return _permute_systems(state, map_original, current_system_dim)


def _permute_systems(mat: torch.Tensor, perm_list: List[int], dim_list: List[int]) -> torch.Tensor:
    if perm_list == list(range(len(dim_list))):
        return mat
    original_shape = mat.shape
    mat = mat.view([-1] + list(mat.shape[-2:]))
    return _permute_dm(mat, perm_list, dim_list).view(original_shape).contiguous()


def _schmidt_decompose(
    psi: torch.Tensor, sys_A: Optional[List[int]] = None  
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

    psi = psi.reshape(-1, 1)
    psi_len = psi.shape[-2]
    # Determine the number of qubits
    num_qubits = int(math.log2(psi_len))

    # If sys_A is not specified, use the first half of qubits
    sys_A = sys_A if sys_A is not None else list(range(num_qubits // 2))
    
    # Determine sys_B as the complement of sys_A
    sys_B = [i for i in range(num_qubits) if i not in sys_A]


    # Permute qubit indices
    psi = psi.reshape([-1] + [2] * num_qubits).permute([0] + [x + 1 for x in sys_A + sys_B])

    # Construct amplitude matrix
    amp_mtr = psi.reshape([-1, 2 ** len(sys_A), 2 ** len(sys_B)])

    # Standard process to obtain Schmidt decomposition
    u, c, v = torch.svd(amp_mtr)

    # Count non-zero singular values
    k = torch.count_nonzero(c > psi_len*1e-13, dim=-1)

    # Select top-k singular values and reshape singular vectors
    c = c[:, :k.max()]
    u = u[:, :, :k.max()].transpose(1, 2).reshape([c.size(0), -1, 1])
    v = v[:, :, :k.max()].transpose(1, 2).reshape([c.size(0), -1, 1])

    return c, u, v



def _adjoint(A, E, f) -> torch.Tensor:
    r"""Calculate gradient of the function f
    Args:
        A: Input matrix.
        E: Undetermined matrix. Usually choose identity
        f: A differentiable function
    Returns:
        The numerical gradient of the function f given input A: \delta f(A)
    """

    batch_dim = list(A.shape[:-2])
    batch_size = [int(np.prod(batch_dim))] 
    A = A.reshape(batch_size + list(A.shape[-2:]))
        
    batch_dims = A.shape[0]
    A_H = A.mH.to(E.dtype)
    n = A.size(1)
    M = torch.zeros(batch_dims, 2 * n, 2 * n, dtype=E.dtype, device=E.device)
    M[:, :n, :n] = A_H
    M[:, n:, n:] = A_H
    M[:, :n, n:] = E
    fM = f(M)
    fM = fM.reshape(batch_size + list(fM.shape[-2:]))
    return fM[:, :n, n:].to(A.dtype).squeeze()



def _logm_scipy(A: torch.Tensor) -> torch.Tensor:
    r"""Calculate logarithm of a matrix
    Args:
        A: Input matrix.
    Returns:
        The matrix of natural base logarithms
    """

    batch_dim = list(A.shape[:-2])
    batch_size = [int(np.prod(batch_dim))] 
    
    A = A.reshape(batch_size + list(A.shape[-2:]))

    batch_dims = A.shape[0]
    logmA = torch.zeros(A.shape, dtype=torch.complex128)
    for i in range(batch_dims):
        logmA[i,:,:] = torch.from_numpy(
            scipy.linalg.logm(A[i,:,:].cpu(), disp=False)[0].astype("complex128"))

    return logmA.squeeze().to(A.device, dtype=A.dtype)



def _perm_of_list(origin: List[int], target: List[int]) -> List[int]:
    r"""Find the permutation mapping the original list to the target list
    """
    perm_map = {val: index for index, val in enumerate(origin)}
    return [perm_map[val] for val in target]


def _permute_sv(state: torch.Tensor, perm: Union[List, Tuple],
                    system_dim: List[int]) -> torch.Tensor:
    r"""speed-up logic using using np.transpose + torch.gather.

    Args:
        state: input state data.
        perm: permutation of system sequence.
        system_dim: list of dimensions for all systems

    Returns:
        permuted state.

    """
    base_idx = torch.arange(int(np.prod(system_dim))).view(system_dim)
    base_idx = torch.permute(base_idx, dims=perm)
    base_idx = base_idx.reshape([1, -1]).expand(state.shape)
    
    return state.gather(1, index=base_idx)


def _permute_dm(state: torch.Tensor, perm: Union[List, Tuple],
                           system_dim: List[int]) -> torch.Tensor:
    r"""speed-up logic using using np.transpose + torch.gather.

    Args:
        state: input state data.
        perm: permutation of system sequence.
        system_dim: list of dimensions for all systems

    Returns:
        permuted state.
    """
    base_idx = torch.arange(int(np.prod(system_dim))).view(system_dim)
    base_idx = torch.permute(base_idx, dims=perm).contiguous()
    
    state = state.gather(1, index=base_idx.view([1, -1, 1]).expand(state.shape))
    state = state.gather(2, index=base_idx.view([1, 1, -1]).expand(state.shape))

    return state

class __Logm(torch.autograd.Function):
    @staticmethod
    def forward(ctx, A):
        assert A.size(-1) == A.size(-2)  # Square matrix
        assert A.dtype in (
            torch.float32,
            torch.float64,
            torch.complex64,
            torch.complex128,
        )
        ctx.save_for_backward(A)
        return _logm_scipy(A)

    @staticmethod
    def backward(ctx, G):
        (A,) = ctx.saved_tensors
        return _adjoint(A, G, _logm_scipy)


_logm = __Logm.apply


def _sqrtm(A: torch.Tensor) -> torch.Tensor:
    w, v = torch.linalg.eigh(A)
    w = torch.clamp(w, min=1e-20)
    sqrt_w = torch.sqrt(w)
    return (v * sqrt_w.unsqueeze(-2)) @ v.mH


def _create_matrix(
    linear_map: Callable[[torch.Tensor], torch.Tensor],
    input_dim: int,
    input_dtype: torch.dtype,
) -> torch.Tensor:
    # Create an identity matrix representing all basis vectors
    identity_matrix = torch.eye(input_dim, dtype=input_dtype)

    # Apply the linear map to each column of the identity matrix
    mapped_vectors = [
        linear_map(identity_matrix[:, i].unsqueeze(1)) for i in range(input_dim)
    ]

    # Stack all mapped vectors to form the final matrix
    return torch.stack(mapped_vectors, dim=1).squeeze()


# TODO: this is a temporal solution for programs that use this function, will be depreciated in the future
def _unitary_transformation(
        U: torch.Tensor, V: torch.Tensor, qubit_idx: Union[List[int], int], num_qubits: int
) -> torch.Tensor:
    r"""Compute :math:`VU`, where :math:`U` is a unitary matrix acting on all qubits.

    Args:
        U: unitary matrix of the Circuit
        V: The gate that acts on the circuit
        qubit_idx: The indices of the qubits on which the gate is acted.
        num_qubits: The number of the qubits in the input quantum state.

    Returns:
        The transformed quantum state.
    """
    # The order of the tensor in paddle is less than 10.
    batch_dims = list(U.shape[:-2])
    num_batch_dims = len(batch_dims)
    
    if not isinstance(qubit_idx, Iterable):
        qubit_idx = [qubit_idx]
        
    # generate swap_list
    num_acted_qubits = len(qubit_idx)
    origin_seq = list(range(num_qubits))
    seq_for_acted = qubit_idx + [x for x in origin_seq if x not in qubit_idx]
    swapped = [False] * num_qubits
    swap_ops = []
    for idx in range(num_qubits):
        if not swapped[idx]:
            next_idx = idx
            swapped[next_idx] = True
            while not swapped[seq_for_acted[next_idx]]:
                swapped[seq_for_acted[next_idx]] = True
                if next_idx < seq_for_acted[next_idx]:
                    swap_ops.append((next_idx, seq_for_acted[next_idx]))
                else:
                    swap_ops.append((seq_for_acted[next_idx], next_idx))
                next_idx = seq_for_acted[next_idx]

    for swap_op in swap_ops:
        shape = batch_dims.copy()
        last_idx = -1
        for idx in swap_op:
            shape.append(2 ** (idx - last_idx - 1))
            shape.append(2)
            last_idx = idx
        shape.append(2 ** (2 * num_qubits - last_idx - 1))
        U = torch.reshape(U, shape)
        U = torch.permute(
            U, tuple(range(num_batch_dims)) + tuple(item + num_batch_dims for item in [0, 3, 2, 1, 4])
        )
    U = torch.reshape(
        U, batch_dims.copy() + [2 ** num_acted_qubits, 2 ** (2 * num_qubits - num_acted_qubits)]
    )
    U = torch.matmul(V, U)
    swap_ops.reverse()
    for swap_op in swap_ops:
        shape = batch_dims.copy()
        last_idx = -1
        for idx in swap_op:
            shape.append(2 ** (idx - last_idx - 1))
            shape.append(2)
            last_idx = idx
        shape.append(2 ** (2 * num_qubits - last_idx - 1))
        U = torch.reshape(U, shape)
        U = torch.permute(
            U, tuple(range(num_batch_dims)) + tuple(item + num_batch_dims for item in [0, 3, 2, 1, 4])
        )
    U = torch.reshape(U, batch_dims.copy() + [2 ** num_qubits, 2 ** num_qubits])
    return U


def _hessian(loss_function: Callable[[torch.Tensor], torch.Tensor], var: torch.Tensor) -> torch.Tensor:
    n, m = var.shape
    mat = torch.empty(m, n, n)
    for i in range(m):
        mat[i, :, :] = torch.autograd.functional.hessian(loss_function, var[:,i])
    return mat


def _gradient(loss_function: Callable[[torch.Tensor], torch.Tensor], var: torch.Tensor, n: int) -> torch.Tensor:
    m = var.shape[0]
    mat = torch.empty(m,1)
    for j in range(m):
        func_var = loss_function(var)
        for _ in range(n):
            grads = torch.autograd.grad(func_var, var, create_graph=True)[0]
            func_var = grads[j]
        mat[j] = func_var
        
    return mat


def _prob_sample(distributions: torch.Tensor, shots: int = 1024, 
                 binary: bool = True, proportional: bool = False) -> Dict[str, torch.Tensor]:
    batch_dim =list(distributions.shape[:-1])
    distributions = distributions.view([-1, distributions.shape[-1]])
    
    num_elements = distributions.size(-1)
    num_bits = num_elements.bit_length() - 1
    
    sampled_indices = torch.multinomial(distributions, shots, replacement=True)
    
    counts = torch.stack([(sampled_indices == i).sum(dim = -1) for i in range(num_elements)], dim = -1)
    
    if proportional:
        counts = counts / shots
    
    keys = [f'{i:0{num_bits}b}' if binary else str(i) for i in range(num_elements)]
    results = OrderedDict((key, counts[:, i].view(batch_dim)) 
                          for i, key in enumerate(keys))
    return dict(results)



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the density_matrix backend.
"""

import math
import warnings
from typing import List, Optional, Tuple, Union

import numpy as np
import torch

from ... import utils
from .. import State


class MixedState(State):
    r"""The mixed state class. 

    Args:
        data: tensor array (in density matrix representation) for quantum mixed state(s).
        sys_dim: a list of dimensions for each system.
        system_seq: the system order of this state. Defaults to be from 1 to n.
        probability: list of state probability distributions. Defaults to be 1.
    
    Note:
        The data is stored in the matrix-form with shape :math:`(-1, d, d)`

    """
    def __init__(self, data: torch.Tensor, sys_dim: List[int], 
                 system_seq: Optional[List[int]] = None,
                 probability: Optional[List[torch.Tensor]] = None):
        dim = int(np.prod(sys_dim))
        self._prob = [] if probability is None else probability
        
        non_batch_len = 2 + len(self._prob)
        self._batch_dim = list(data.shape[:-non_batch_len])

        data = data.reshape([-1, dim, dim])
        super().__init__(data, sys_dim, system_seq)
    
    def __getitem__(self, key: Union[int, slice]) -> 'MixedState':
        assert self.batch_dim, \
            f"This state is not batched and hence cannot be indexed: received key {key}."
        return MixedState(self.density_matrix[key], self.system_dim, self.system_seq,
                          [prob.clone() for prob in self._prob])
    
    def prob_select(self, outcome_idx: torch.Tensor, prob_idx: int = -1) -> 'MixedState':
        num_prob = len(self._prob)
        if prob_idx > 0:
            prob_idx -= len(self._prob) 

        new_prob = []
        for idx, prob in enumerate(self._prob):
            if num_prob + prob_idx > idx:
                new_prob.append(prob.clone())
            else:
                new_prob.append(prob.index_select(dim=prob_idx, index=outcome_idx).squeeze(prob_idx))

        data_idx = prob_idx - 2
        data = self.density_matrix.index_select(dim=data_idx, index=outcome_idx).squeeze(data_idx)
        return MixedState(data, self.system_dim, self.system_seq, new_prob)
    
    def expand(self, batch_dim: List[int]) -> 'MixedState':
        if self._prob != []:
            raise NotImplementedError(
                "Indexing of mixed state with probabilistic computation is not supported."
            )
        
        expand_state = self.clone()
        expand_state._batch_dim = batch_dim
        expand_state._prob = []
        
        if np.prod(batch_dim) == np.prod(self._batch_dim):
            return expand_state
        
        non_batch_len = 2 + len(self._prob)
        expand_state._data = self._data.expand(batch_dim + [-1] * non_batch_len).reshape([-1, self.dim, self.dim])
        return expand_state
    
    @property
    def backend(self) -> str:
        return "density_matrix"
    
    @staticmethod
    def check(data: torch.Tensor, sys_dim: Union[int, List[int]], eps: Optional[float] = 1e-4) -> int:
        if isinstance(sys_dim, List):
            num_sys = len(sys_dim)
            expected_dimension = int(np.prod(sys_dim))
        else:
            num_sys = int(math.log(data.shape[-2], sys_dim))
            expected_dimension = sys_dim ** num_sys
        
        assert data.shape[-2] == expected_dimension, \
            f"Input data shape {data.shape} does not match the input system dimension {sys_dim}."
        
        if eps:
            assert torch.all(is_density := utils.check._is_density_matrix(data, eps)), \
                f"(Some) data is not a density matrix: asserted {is_density.tolist()}"
        
        return num_sys

    @property
    def ket(self) -> torch.Tensor:
        raise NotImplementedError(
            "Mixed state does not support state vector representation." +
            "If you are looking for the vectorization of the mixed state, please call the 'vec' property.")

    @property
    def density_matrix(self) -> torch.Tensor:
        self.reset_sequence()
        return self._data.view(self.batch_dim + [self.dim, self.dim]).clone()
    
    def _trace(self, trace_idx: List[int]) -> 'MixedState':
        remain_seq, remain_system_dim = [], []
        for x in self._system_seq:
            if x not in trace_idx:
                remain_seq.append(x)
                remain_system_dim.append(self.system_dim[x])
        trace_dim = math.prod([self.system_dim[x] for x in trace_idx])
        self.system_seq = trace_idx + remain_seq
        
        data = self._data.view(self.batch_dim + [self.dim, self.dim])
        data = utils.linalg._trace_1(data, trace_dim)
        
        # convert remaining sequence
        value_to_index = {value: index for index, value in enumerate(sorted(remain_seq))}
        remain_seq = [value_to_index[i] for i in remain_seq]
        return MixedState(data, remain_system_dim, remain_seq, self._prob)
    
    def _transpose(self, transpose_idx: List[int]) -> 'MixedState':
        self.system_seq = transpose_idx + [x for x in self._system_seq if x not in transpose_idx]
        transpose_dim = math.prod([self.system_dim[x] for x in transpose_idx])
        
        state = self.clone()
        state._data = utils.linalg._transpose_1(state._data, transpose_dim)
        return state

    def normalize(self) -> None:
        self._data = torch.div(self._data, utils.linalg._trace(self._data, -2, -1).view([-1, 1, 1]))

    def clone(self) -> 'MixedState':
        dim = self.dim
        data = self._data.view(self.batch_dim + [dim, dim]).clone()
        return MixedState(data, self.system_dim, self.system_seq, [prob.clone() for prob in self._prob])

    def fit(self, backend: str) -> torch.Tensor:
        data = self.density_matrix
        if backend == self.backend:
            return data
        
        if backend == 'state_vector':
            # TODO: add batch support
            return utils.linalg._density_to_vector(data)
        
        raise NotImplementedError(
            f"Unsupported state conversion from 'density_matrix' to {backend}.")

    @State.system_seq.setter
    def system_seq(self, target_seq: List[int]) -> None:
        if target_seq == self._system_seq:
            return
        
        perm_map = utils.linalg._perm_of_list(self._system_seq, target_seq)
        current_system_dim = [self._sys_dim[x] for x in self._system_seq]

        self._data = utils.linalg._permute_dm(self._data, perm_map, current_system_dim).contiguous()
        self._system_seq = target_seq
    
    def _evolve(self, unitary: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        dim, _shape = self.dim, self._squeeze_shape
        if on_batch:
            self._batch_dim = self._batch_dim or list(unitary.shape[:-2])
            evolve_axis = [-1, 1]
        else:
            evolve_axis = [1, -1]
        
        applied_dim = unitary.shape[-1]
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        unitary = unitary.view(evolve_axis + [applied_dim, applied_dim])
        data = self._data.view(_shape + [applied_dim, (dim ** 2) // applied_dim])
        data = torch.matmul(unitary, data)
        
        data = data.view(_shape + [dim, applied_dim, dim // applied_dim])
        self._data = torch.matmul(unitary.unsqueeze(-3).conj(), data).view([-1, dim, dim])
    
    def _evolve_ctrl(self, unitary: torch.Tensor, index: int, 
                     sys_idx: List[Union[int, List[int]]], on_batch: bool = True) -> None:
        dim, _shape = self.dim, self._squeeze_shape
        if on_batch:
            self._batch_dim = self._batch_dim or list(unitary.shape[:-2])
            evolve_axis = [-1, 1]
        else:
            evolve_axis = [1, -1]
        
        ctrl_idx = sys_idx[0]
        ctrl_dim = np.prod([self.system_dim[idx] for idx in ctrl_idx])
        
        sys_idx = ctrl_idx + sys_idx[1:]
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        applied_dim = unitary.shape[-1]
        unitary = unitary.view(evolve_axis + [applied_dim, applied_dim])
        
        other_dim = dim // ctrl_dim
        data = self._data.squeeze().expand(self.batch_dim + [dim, dim])
        
        data = data.view(_shape + [ctrl_dim, applied_dim, (other_dim ** 2) * ctrl_dim // applied_dim]).clone()
        data[:, :, index] = torch.matmul(unitary, data[:, :, index])
        
        data = data.view(_shape + [dim, ctrl_dim, applied_dim, other_dim // applied_dim]).clone()
        data[:, :, :, index] = torch.matmul(unitary.unsqueeze(-3).conj(), data[:, :, :, index])
        
        self._data = data.view([-1, dim, dim])
    
    def _evolve_keep_dim(self, unitary: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        unitary_batch_dim = list(unitary.shape[:-2])
        dim, _shape = self.dim, self._squeeze_shape
        
        if on_batch:
            self._batch_dim = (self._batch_dim or unitary_batch_dim[:-1]) + unitary_batch_dim[-1:]
            evolve_axis = [-1, 1]
        else:
            evolve_axis = [1, -1]
        
        applied_dim, num_unitary = unitary.shape[-1], int(np.prod(unitary_batch_dim[-1:]))
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        unitary = unitary.view(evolve_axis + [num_unitary, applied_dim, applied_dim])
        data = self._data.view(_shape + [1, applied_dim, (dim ** 2) // applied_dim])
        data = torch.matmul(unitary, data)
        
        data = data.view(_shape + [num_unitary, dim, applied_dim, dim // applied_dim])
        unitary = unitary.unsqueeze(-3)
        data = torch.matmul(unitary.conj(), data)
        self._data = data.view([-1, dim, dim])
    
    def _expec_val(self, obs: torch.Tensor, sys_idx: List[int]) -> torch.Tensor:
        dim = self.dim
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        applied_dim, num_obs = obs.shape[-1], obs.shape[-3]
        state = self._data.view([-1, 1, applied_dim, (dim ** 2) // applied_dim])
        state = torch.matmul(obs, state).view([-1, dim, dim])
    
        return utils.linalg._trace(state, -2, -1).view(self.batch_dim + [num_obs])
    
    def _expec_state(self, prob_idx: List[int]) -> 'MixedState':
        dim, num_prob = self.dim, len(self._prob)
        batch_prob_len = len(self._prob[-1].shape[:-num_prob])
        prob = self._joint_probability(prob_idx)
        
        states = self._data.view([-1] + self._prob_dim + [dim ** 2])
        prob = prob.view(list(prob.shape) + [1] * (self._prob[-1].dim() - prob.dim() + 1))
        prob_state = torch.mul(prob, states)

        sum_idx = [idx + 1 for idx in prob_idx]
        expectation = prob_state.sum(sum_idx)

        new_prob = []
        if len(prob_idx) != num_prob:
            new_prob = [prob.clone() for idx, prob in enumerate(self._prob) if idx not in prob_idx]
            expectation = expectation.view(self._batch_dim + list(new_prob[-1].shape[batch_prob_len:]) + [dim, dim])
        else:
            expectation = expectation.view(self._batch_dim + [dim, dim])
        return MixedState(expectation, self.system_dim, self.system_seq, new_prob)
    
    def _measure(self, measure_op: torch.Tensor, sys_idx: List[int]) -> Tuple[torch.Tensor, 'MixedState']:
        new_state = self.clone()
        new_state._evolve_keep_dim(measure_op, sys_idx)
        
        data = new_state._data
        measure_prob = utils.linalg._trace(data, -2, -1).real.view([-1, 1, 1])
        collapsed_data = data / measure_prob
        collapsed_data[collapsed_data != collapsed_data] = 0
        
        measure_prob = measure_prob.view(new_state._batch_dim[:-1] + self._prob_dim + [-1])
        new_state._data = collapsed_data
        new_state._batch_dim = new_state._batch_dim[:-1]
        new_state._prob.append(measure_prob)
        return measure_prob, new_state
    
    def __kraus_transform(self, list_kraus: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        r"""Apply the Kraus operator to the state.
        
        Args:
            list_kraus: the Kraus operators.
            sys_idx: the system index list.
            on_batch: whether this operator evolves on batch axis. Defaults to True.
        
        """
        self._evolve_keep_dim(list_kraus, sys_idx)
        self._batch_dim, rank = self._batch_dim[:-1], self._batch_dim[-1]
        self._data = self._data.view([-1, rank, self.dim, self.dim]).sum(dim=-3)
    
    def __choi_transform(self, choi: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        r"""Apply the Choi operator to the state.
        
        Args:
            choi: the Choi operator.
            sys_idx: the system index list.
            on_batch: whether this operator evolves on batch axis. Defaults to True.
        
        """
        _shape = self._squeeze_shape
        if on_batch:
            self._batch_dim = self._batch_dim or list(choi.shape[:-2])
            transform_axis = [-1, 1]
        else:
            transform_axis = [1, -1]
        
        refer_sys_idx = [x for x in self._system_seq if x not in sys_idx]
        dim_refer = int(np.prod([self.system_dim[x] for x in refer_sys_idx]))
        dim_in = int(np.prod([self.system_dim[x] for x in sys_idx]))
        dim_out = choi.shape[-1] // dim_in
        
        self.system_seq = refer_sys_idx + sys_idx
        data = self._data.view(_shape + [dim_refer, dim_in, dim_refer, dim_in])
        choi = choi.view(transform_axis + [dim_in, dim_in * (dim_out ** 2)])
        
        data = torch.matmul(data.transpose(-1, -3).reshape(_shape + [(dim_refer ** 2) * dim_in, dim_in]), choi)
        data = torch.transpose(data.view(_shape + [dim_in, dim_refer, dim_out, dim_in, dim_out]), -3, -4)
        data = utils.linalg._trace(data, -2, -5)
        self._data = data.view([-1, dim_refer * dim_out, dim_refer * dim_out])
    
    def _transform(self, op: torch.Tensor, sys_idx: List[int], repr_type: str) -> None:
        self.__kraus_transform(op, sys_idx) if repr_type == 'kraus' else self.__choi_transform(op, sys_idx)
    
    def sqrt(self) -> torch.Tensor:
        return utils.linalg._sqrtm(self.density_matrix)
    
    def log(self) -> torch.Tensor:
        if self.rank < self.dim:
            warnings.warn(
                f"The matrix logarithm may not be accurate: expect rank {self.dim}, received {self.rank}", UserWarning)
        return utils.linalg._logm(self.density_matrix)



# !/usr/bin/env python3
# Copyright (c) 2024 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import math
from functools import reduce
from typing import List, Union

import numpy as np
import torch
from torch.linalg import matrix_power

from .. import database
from ..core import State, get_dtype, to_state, utils
from ..core.intrinsic import _alias

__all__ = [
    "pauli_basis",
    "pauli_group",
    "pauli_str_basis",
    "pauli_str_povm",
    "qft_basis",
    "std_basis",
    "bell_basis",
    "heisenberg_weyl",
    "phase_space_point",
    "gell_mann",
]


def pauli_basis(num_qubits: int) -> torch.Tensor:
    r"""Generate a Pauli basis.

    Args:
        num_qubits: the number of qubits :math:`n`.

    Returns:
         The Pauli basis of :math:`\mathbb{C}^{2^n \times 2^n}`, where each tensor is accessible along the first dimension.

    Examples:
        .. code-block:: python

            num_qubits = 1
            basis = pauli_basis(num_qubits)
            print(f'The Pauli basis is:\n{basis}')

        ::

            The Pauli basis is:
            tensor([[[ 0.7071+0.0000j,  0.0000+0.0000j],
                     [ 0.0000+0.0000j,  0.7071+0.0000j]],

                    [[ 0.0000+0.0000j,  0.7071+0.0000j],
                     [ 0.7071+0.0000j,  0.0000+0.0000j]],

                    [[ 0.0000+0.0000j,  0.0000-0.7071j],
                     [ 0.0000+0.7071j,  0.0000+0.0000j]],

                    [[ 0.7071+0.0000j,  0.0000+0.0000j],
                     [ 0.0000+0.0000j, -0.7071+0.0000j]]])
    """
    single_pauli_basis = torch.stack([database.matrix.eye(), 
                                      database.matrix.x(), 
                                      database.matrix.y(), 
                                      database.matrix.z()]) * math.sqrt(2) / 2
    if num_qubits == 1:
        return single_pauli_basis
    return reduce(
        lambda result, index: torch.kron(result, index),
        [single_pauli_basis for _ in range(num_qubits - 2)],
        torch.kron(single_pauli_basis, single_pauli_basis),
    )


def pauli_group(num_qubits: int) -> torch.Tensor:
    r"""Generate a Pauli group i.e., an unnormalized Pauli basis.

    Args:
        num_qubits: the number of qubits :math:`n`.

    Returns:
         The Pauli group of :math:`\mathbb{C}^{2^n \times 2^n}`, where each tensor is accessible along the first dimension.

    Examples:
        .. code-block:: python

            num_qubits = 1
            group = pauli_group(num_qubits)
            print(f'The Pauli group is:\n{group}')

        ::

            The Pauli group is:
            tensor([[[ 1.0000+0.0000j,  0.0000+0.0000j],
                     [ 0.0000+0.0000j,  1.0000+0.0000j]],

                    [[ 0.0000+0.0000j,  1.0000+0.0000j],
                     [ 1.0000+0.0000j,  0.0000+0.0000j]],

                    [[ 0.0000+0.0000j,  0.0000-1.0000j],
                     [ 0.0000+1.0000j,  0.0000+0.0000j]],

                    [[ 1.0000+0.0000j,  0.0000+0.0000j],
                     [ 0.0000+0.0000j, -1.0000+0.0000j]]])
    """
    return pauli_basis(num_qubits) * (math.sqrt(2) ** num_qubits)


def pauli_str_basis(pauli_str: Union[str, List[str]]) -> State:
    r"""Get the state basis with respect to the Pauli string.
    
    Args:
        pauli_str: the string composed of 'i', 'x', 'y' and 'z' only.
    
    Returns:
        The state basis of the observable given by the Pauli string.
        
    Examples:
        .. code-block:: python

            pauli_str = ['x', 'z']
            state_basis = pauli_str_basis(pauli_str)
            print(f'The state basis of the observable is:\n{state_basis}')

        ::

            The state basis of the observable is:
            ---------------------------------------------------
            Backend: state_vector
            System dimension: [2]
            System sequence: [0]
            Batch size: [2, 2]

            # 0:
            [0.71+0.j 0.71+0.j]
            # 1:
            [ 0.71+0.j -0.71+0.j]
            # 2:
            [1.+0.j 0.+0.j]
            # 3:
            [0.+0.j 1.+0.j]
            ---------------------------------------------------
    """
    x_basis = torch.tensor([[1, 1], 
                            [1, -1]], dtype=torch.complex128) / math.sqrt(2)
    y_basis = torch.tensor([[1, 1j],
                            [1, -1j]], dtype=torch.complex128) / math.sqrt(2)
    z_basis = torch.eye(2, dtype=torch.complex128)
    i_basis = z_basis
    locals_ = locals()
    
    def __get_single_str(string: str) -> torch.Tensor:
        list_basis = [locals_[f'{p}_basis'] for p in string.lower()]
        basis = utils.linalg._nkron(*list_basis) if len(list_basis) > 1 else list_basis[0]
        return basis.unsqueeze(-1)
    
    basis = __get_single_str(pauli_str) if isinstance(pauli_str, str) else torch.stack([__get_single_str(string) for string in pauli_str])
    return to_state(basis.to(dtype=get_dtype()))


def pauli_str_povm(pauli_str: Union[str, List[str]]) -> torch.Tensor:
    r"""Get the povm with respect to the Pauli string.
    
    Args:
        pauli_str: the string composed of 'i', 'x', 'y' and 'z' only.
    
    Returns:
        The POVM of the observable given by the Pauli string.
    
    Examples:
        .. code-block:: python

            pauli_str = ['x', 'y']
            POVM = pauli_str_povm(pauli_str)
            print(f'The POVM of the observable is:\n{POVM}')

        ::

            The POVM of the observable is:
            tensor([[[[ 0.5000+0.0000j,  0.5000+0.0000j],
                      [ 0.5000+0.0000j,  0.5000+0.0000j]],

                     [[ 0.5000+0.0000j, -0.5000+0.0000j],
                      [-0.5000+0.0000j,  0.5000+0.0000j]]],


                    [[[ 0.5000+0.0000j,  0.0000-0.5000j],
                      [ 0.0000+0.5000j,  0.5000+0.0000j]],

                     [[ 0.5000+0.0000j,  0.0000+0.5000j],
                      [ 0.0000-0.5000j,  0.5000+0.0000j]]]])
    """
    return pauli_str_basis(pauli_str).density_matrix


def qft_basis(num_qubits: int) -> State:
    r"""Compute the eigenvectors (eigenbasis) of the Quantum Fourier Transform (QFT) matrix.

    Args:
        num_qubits: Number of qubits :math:`n` such that :math:`N = 2^n`.

    Returns:
        A tensor where the first index gives the eigenvector of the QFT matrix.
    
    Examples:
        .. code-block:: python

            num_qubits = 2
            qft_state = qft_basis(num_qubits)
            print(f'The eigenvectors of the QFT matrix is:\n{qft_state}')

        ::

            The eigenvectors of the QFT matrix is:
            ---------------------------------------------------
            Backend: state_vector
            System dimension: [2]
            System sequence: [0]
            Batch size: [2]

            # 0:
            [0.92+0.j 0.38+0.j]
            # 1:
            [-0.38-0.j  0.92+0.j]
            ---------------------------------------------------
    """
    #TODO numerically unstable, needs a more precise implementation instead of using eig decomposition
    _, eigvec = torch.linalg.eig(database.matrix.qft_matrix(num_qubits))
    return to_state(eigvec.T.unsqueeze(-1).to(get_dtype()))


@_alias({"num_systems": "num_qubits"})
def std_basis(num_systems: int, system_dim: Union[List[int], int] = 2) -> State:
    r"""Generate all standard basis states for a given number of qubits.

    Args:
        num_systems: number of systems in this state. Alias of ``num_qubits``.
        system_dim: dimension of systems. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to be qubit case.

    Returns:
        A tensor where the first index gives the computational vector.
    
    Examples:
        .. code-block:: python

            num_systems = 2
            system_dim = [1, 2]
            basis = std_basis(num_systems, system_dim)
            print(f'The standard basis states are:\n{basis}')

        ::

            The standard basis states are:
            ---------------------------------------------------
            Backend: state_vector
            System dimension: [1, 2]
            System sequence: [0, 1]
            Batch size: [2]

            # 0:
            [1.+0.j 0.+0.j]
            # 1:
            [0.+0.j 1.+0.j]
            ---------------------------------------------------
    """
    dim = system_dim ** num_systems if isinstance(system_dim, int) else math.prod(system_dim)
    return to_state(torch.eye(dim).unsqueeze(-1).to(get_dtype()), system_dim)


def bell_basis() -> State:
    r"""Generate the Bell basis for a 2-qubit system, 
    with each basis state accessible along the first dimension of a tensor.

    Returns:
        A tensor of shape (4, 4, 1), representing the four Bell basis states.
    
    Examples:
        .. code-block:: python

            basis = bell_basis()
            print(f'The Bell basis for a 2-qubit system are:\n{basis}')

        ::

            The Bell basis for a 2-qubit system are:
            ---------------------------------------------------
            Backend: state_vector
            System dimension: [2, 2]
            System sequence: [0, 1]
            Batch size: [4]

            # 0:
            [0.71+0.j 0.  +0.j 0.  +0.j 0.71+0.j]
            # 1:
            [ 0.71+0.j  0.  +0.j  0.  +0.j -0.71+0.j]
            # 2:
            [0.  +0.j 0.71+0.j 0.71+0.j 0.  +0.j]
            # 3:
            [ 0.  +0.j  0.71+0.j -0.71+0.j  0.  +0.j]
            ---------------------------------------------------
    """
    mat = torch.tensor([
        [ 1,  0,  0,  1],   # |Φ+⟩ = (|00⟩ + |11⟩) / √2
        [ 1,  0,  0, -1],   # |Φ-⟩ = (|00⟩ - |11⟩) / √2
        [ 0,  1,  1,  0],   # |Ψ+⟩ = (|01⟩ + |10⟩) / √2
        [ 0,  1, -1,  0]    # |Ψ-⟩ = (|01⟩ - |10⟩) / √2
    ], dtype=torch.complex128) / math.sqrt(2)
    return to_state(mat.unsqueeze(-1).to(get_dtype()))


def heisenberg_weyl(dim: int) -> torch.Tensor:
    r"""Generate Heisenberg-Weyl operator for qudit. 
    The Heisenberg-Weyl operators are defined as :math:`T(a,b) = e^{-(d+1) \pi i a b/ d}Z^a X^b`.

    Args:
        dim: dimension of qudit

    Returns:
        Heisenberg-Weyl operator for qudit.
        
    Examples:
        .. code-block:: python

            dim = 2
            operator = heisenberg_weyl(dim)
            print(f'The Heisenberg-Weyl operator for qudit is:\n{operator}')

        ::

            The Heisenberg-Weyl operator for qudit is:
            tensor([[[ 1.0000e+00+0.0000e+00j,  0.0000e+00+0.0000e+00j],
                     [ 0.0000e+00+0.0000e+00j,  1.0000e+00+0.0000e+00j]],

                    [[ 1.0000e+00+0.0000e+00j,  0.0000e+00+0.0000e+00j],
                     [ 0.0000e+00+0.0000e+00j, -1.0000e+00+1.2246e-16j]],

                    [[ 0.0000e+00+0.0000e+00j,  1.0000e+00+0.0000e+00j],
                     [ 1.0000e+00+0.0000e+00j,  0.0000e+00+0.0000e+00j]],

                    [[-0.0000e+00+0.0000e+00j, -1.8370e-16+1.0000e+00j],
                     [ 6.1232e-17-1.0000e+00j, -0.0000e+00+0.0000e+00j]]])
    """
    complex_dtype = get_dtype() 
    _phase = database.matrix.phase(dim)
    _shift = database.matrix.shift(dim)

    phase_pow = torch.zeros(dim, 1, dim, dim, dtype=complex_dtype)
    shift_pow = torch.zeros(1, dim, dim, dim, dtype=complex_dtype)

    a = torch.arange(dim, dtype=torch.float64).reshape(1, dim, 1, 1)
    b = torch.arange(dim, dtype=torch.float64).reshape(dim, 1, 1, 1)

    for i in range(dim):
        phase_pow[i, 0, :, :] = matrix_power(_phase, i)
        shift_pow[0, i, :, :] = matrix_power(_shift, i)

    hw = torch.matmul(phase_pow, shift_pow)
    ab = torch.matmul(a, b)
    ab = torch.exp(-ab * 1j * np.pi * (dim + 1) / dim)

    hw = ab * hw
    hw = hw.permute(1, 0, 2, 3)
    hw = hw.reshape((dim**2, dim, dim))
    return hw


def phase_space_point(dim: int) -> torch.Tensor:
    r"""Generate phase space point operator for qudit.

    Args:
        dim: dimension of qudit

    Returns:
        Phase space point operator for qudit.
        
    Examples:
        .. code-block:: python

            dim = 2
            operator = phase_space_point(dim)
            print(f'The phase space point operator for qudit is:\n{operator}')

        ::

            The phase space point operator for qudit is:
            tensor([[[ 1.0000+0.0000e+00j,  0.5000+5.0000e-01j],
                     [ 0.5000-5.0000e-01j,  0.0000+6.1232e-17j]],

                    [[ 1.0000+0.0000e+00j, -0.5000-5.0000e-01j],
                     [-0.5000+5.0000e-01j,  0.0000+6.1232e-17j]],

                    [[ 0.0000+6.1232e-17j,  0.5000-5.0000e-01j],
                     [ 0.5000+5.0000e-01j,  1.0000+0.0000e+00j]],

                    [[ 0.0000+6.1232e-17j, -0.5000+5.0000e-01j],
                     [-0.5000-5.0000e-01j,  1.0000+0.0000e+00j]]])
    """

    hw = heisenberg_weyl(dim)
    A0 = torch.sum(hw, dim=0) / dim

    hw_dagger = hw.mH
    A = A0.expand(hw.shape[0], A0.shape[0], A0.shape[1])
    A = torch.matmul(hw, A)
    A = torch.matmul(A, hw_dagger)
        
    return A

def __gell_mann(index1: int, index2: int, dim: int) -> torch.Tensor:
    if index1 == index2:
        if index1 == 0:
            mat = torch.eye(dim)
        else:
            N = math.sqrt(2 / (index1 * (index1 + 1)))
            
            diag_element = torch.zeros(dim)
            diag_element[:index1] = 1
            diag_element[index1] = -index1
            
            mat = N * torch.diag(diag_element).to(torch.complex128)
    else:
        # Off-diagonal elements
        E = torch.zeros(dim, dim)
        E[index1, index2] = 1
        mat = E + E.T if index1 < index2 else 1j * (E - E.T)
    return mat

def gell_mann(dim: int) -> torch.Tensor:
    r"""Generate a set of Gell-Mann matrices for a given dimension. These matrices span the entire space 
    of dim-by-dim matrices, and they generalize the Pauli operators when dim = 2 and the Gell-Mann operators 
    when dim = 3.

    Args:
        dim: a positive integer indicating the dimension.

    Returns:
        A set of Gell-Mann matrices.
        
    Examples:
        .. code-block:: python

            dim = 2
            matrices = gell_mann(dim)
            print(f'The Gell-Mann matrices are:\n{matrices}')

        ::

            The Gell-Mann matrices are:
            tensor([[[ 0.+0.j,  1.+0.j],
                     [ 1.+0.j,  0.+0.j]],

                    [[ 0.+0.j, -0.-1.j],
                     [ 0.+1.j,  0.+0.j]],

                    [[ 1.+0.j,  0.+0.j],
                     [ 0.+0.j, -1.+0.j]]])
    """
    list_gell_mann = [
        __gell_mann(idx1, idx2, dim).unsqueeze(0)
        for idx1, idx2 in itertools.product(range(dim), range(dim))
        if idx1 != 0 or idx2 != 0
    ]
    list_gell_mann = torch.cat(list_gell_mann, dim=0)
    return list_gell_mann.to(get_dtype())



# !/usr/bin/env python3
# Copyright (c) 2024 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the PQCombNet class.
"""

import csv
import itertools
import os
import time
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from torch.nn import ModuleList

from quairkit.circuit import Circuit
from quairkit.core import set_seed, to_state, utils
from quairkit.database import bell_state, random_unitary, zero_state
from quairkit.operator import ParamOracle
from quairkit.qinfo import channel_repr_convert

__all__ = ["PQCombNet"]


class PQCombNet:
    r"""
    Parameterized Quantum Comb Net.

    Args:
        target_function: The function to apply to each unitary in the dataset.
        num_slots: The number of unitaries to be queried.
        ancilla: The ancilla dimension or dimension list.
        slot_dim: The slot dimension for the unitaries to be queried.
        train_unitary_info: The number of unitaries or the unitary dataset to be used for training or the training unitary set.
        test_unitary_info: The number of unitaries or the unitary dataset to be used for testing or the testing unitary set.
        train_mode: The training mode, which can be "process", "comb", or "swap", default is "process".
        task_name: Optional name for the task, useful for data logging and storage.
        is_ctrl_U: Flag to indicate if a controlled-U operation is used in the training process.
        seed: Optional seed for random number generation, enhancing reproducibility.
    """

    def __init__(
        self,
        target_function: Callable[[torch.Tensor], torch.Tensor],
        num_slots: int = 1,
        ancilla: Union[List[int], int] = 0,
        slot_dim: int = 2,
        train_unitary_info: Union[int, torch.Tensor] = 1000,
        test_unitary_info: Union[int, torch.Tensor] = 0,
        train_mode: str = "process",
        task_name: str = None,
        is_ctrl_U: bool = False,
        seed: Optional[int] = None,
    ) -> None:
        _setup_parameters(
            self,
            target_function,
            num_slots,
            ancilla,
            slot_dim,
            train_mode,
            task_name,
            is_ctrl_U,
            seed,
        )
        _setup_unitary_sets(self, train_unitary_info, test_unitary_info)
        _initialize_training_environment(self)

    @property
    def target_function(self) -> Callable[[torch.Tensor], torch.Tensor]:
        return self._target_function

    @target_function.setter
    def target_function(self, value: Callable[[torch.Tensor], torch.Tensor]) -> None:
        self._target_function = value
        _calculate_omegas(self)

    @property
    def num_slots(self) -> int:
        return self._num_slots

    @num_slots.setter
    def num_slots(self, value: int) -> None:
        self._num_slots = value
        self._num_V = self._num_slots + 1
        if self._train_mode == "comb":
            _calculate_omegas(self)

    @property
    def ancilla_dim_list(self) -> List[int]:
        return self._ancilla_dim_list

    @ancilla_dim_list.setter
    def ancilla_dim_list(self, value: List[int]) -> None:
        self._ancilla_dim_list = value
        self._system_dim_list = self._ancilla_dim_list + [self.slot_dim]

    @property
    def ancilla_dim(self) -> int:
        return np.prod(self.ancilla_dim_list).__int__()

    @property
    def slot_dim(self) -> int:
        return self._slot_dim

    @property
    def train_mode(self) -> str:
        return self._train_mode

    @train_mode.setter
    def train_mode(self, value: str) -> None:
        if self.is_ctrl_U:
            raise ValueError("Cannot set train_mode when is_ctrl_U is True")
        self._train_mode = value

    @property
    def LR(self) -> float:
        return self._LR

    @LR.setter
    def LR(self, value: float) -> None:
        self._LR = value

    @property
    def NUM_ITR(self) -> int:
        return self._NUM_ITR

    @property
    def task_name(self) -> str:
        return self._task_name

    @property
    def seed(self) -> int:
        return self._seed

    @seed.setter
    def seed(self, value: int) -> None:
        self._seed = value
        set_seed(self._seed)

    @property
    def is_ctrl_U(self) -> bool:
        return self._is_ctrl_U

    @property
    def num_V(self) -> int:
        return self._num_V

    @property
    def train_unitary_set(self) -> torch.Tensor:
        return self._train_unitary_set

    @train_unitary_set.setter
    def train_unitary_set(self, value: torch.Tensor) -> None:
        self._train_unitary_set = value
        _calculate_omegas(self)

    @property
    def test_unitary_set(self) -> torch.Tensor:
        return self._test_unitary_set

    @test_unitary_set.setter
    def test_unitary_set(self, value: torch.Tensor) -> None:
        self._test_unitary_set = value
        _calculate_omegas(self)

    @property
    def omega_train(self) -> torch.Tensor:
        return self._omega_train

    @property
    def omega_test(self) -> torch.Tensor:
        return self._omega_test

    @property
    def V_circuit_list(self) -> ModuleList:
        return self._V_circuit_list

    @V_circuit_list.setter
    def V_circuit_list(self, value: ModuleList) -> None:
        self.num_slots = value.__len__() - 1
        self.ancilla_dim_list = value[0].system_dim[:-1]
        self._V_circuit_list = value

    @property
    def data_directory_name(self) -> str:
        return self._data_directory_name

    @data_directory_name.setter
    def data_directory_name(self, value: str) -> None:
        self._data_directory_name = value

    @property
    def system_dim_list(self) -> List[int]:
        return self._system_dim_list

    @property
    def system_dim(self) -> int:
        return np.prod(self.system_dim_list).__int__()

    def update_V_circuit(
        self,
        index: int,
        new_V: Union[
            Circuit, ParamOracle, torch.Tensor, Tuple[torch.Tensor, List[int]]
        ],
    ) -> None:
        r"""
        Update the V circuit at the specified index with a new circuit.

        Args:
            index: The index of the V circuit to update.
            new_V: The new V circuit, which can be a ParamOracle, Circuit, torch.Tensor or Tuple[torch.Tensor, List[int]].

        Raises:
            ValueError: If the index is out of range or if the dimension of the provided Circuit does not match the dimension of the existing Circuit.
            TypeError: If the new_V is not a Circuit, ParamOracle, torch.Tensor, or Tuple[torch.Tensor, List[int]].
        """
        if not (0 <= index < len(self.V_circuit_list)):
            raise ValueError(f"Index out of range: {index}")

        if isinstance(V := new_V, Circuit):
            if new_V.system_dim != self.system_dim_list:
                raise ValueError(
                    f"The dimension of the provided Circuit does not match the dimension of the existing Circuit: {new_V.system_dim} != {self.system_dim_list}"
                )
        else:
            V = Circuit(system_dim=self.system_dim_list)
            if isinstance(new_V, ParamOracle):
                V.append(new_V)
            elif isinstance(new_V, torch.Tensor):
                V.oracle(oracle=new_V, system_idx=self.system_dim_list)
            elif (
                isinstance(new_V, tuple)
                and isinstance(new_V[0], torch.Tensor)
                and isinstance(new_V[1], list)
            ):
                V.oracle(oracle=new_V[0], system_idx=new_V[1])
            else:
                raise TypeError(
                    "new_V must be a Circuit, ParamOracle, torch.Tensor, Tuple[torch.Tensor, or List[int]]"
                )

        self.V_circuit_list[index] = V

    def train(
        self,
        projector: Optional[torch.Tensor] = None,
        base_lr: float = 0.1,
        max_epochs: int = 10000,
        is_save_data: bool = False,
        is_auto_stop: bool = True,
    ) -> None:
        r"""
        Train the PQCombNet model.

        Args:
            projector: The projector to apply to the ancilla system of the output state.
            base_lr: The base learning rate for the optimizer.
            max_epochs: The maximum number of epochs to train for.
            is_save_data: A flag to indicate whether to save the training data.
            is_auto_stop: A flag to indicate whether to stop training early if the learning rate is too low.
        """
        if self.train_mode == "swap":
            _swap_train(
                self, projector, base_lr, max_epochs, is_save_data, is_auto_stop
            )
        else:
            _train(self, projector, base_lr, max_epochs, is_save_data, is_auto_stop)

    def extract_highest_fidelity(self) -> None:
        r"""
        Call the _extract_highest_fidelity function to generate the fidelity tables.
        If the file does not exist, prompt the user to set is_save_data to True.
        """
        filepath = os.path.join(
            self.data_directory_name, f"{self.task_name}_train_log.csv"
        )
        if not os.path.exists(filepath):
            print(
                f"File {filepath} does not exist. Consider setting is_save_data to True."
            )
            return

        _extract_highest_fidelity(
            self.data_directory_name, f"{self.task_name}_train_log.csv"
        )

    def plot(self):
        r"""
        Plot the quantum comb circuit.
        """
        cir = Circuit(system_dim=self.system_dim_list)
        for index, V_circuit in enumerate(self.V_circuit_list):
            cir.extend(deepcopy(V_circuit))
            if index < self.num_slots:
                if self.is_ctrl_U:
                    cir.control_oracle(
                        torch.eye(self.slot_dim),
                        [len(self.ancilla_dim_list) - 1, len(self.ancilla_dim_list)],
                        latex_name=(r"U" if index % 2 == 0 else r"U^{\dagger}"),
                    )
                else:
                    cir.oracle(
                        torch.eye(self.slot_dim),
                        len(self.ancilla_dim_list),
                        latex_name=(r"U"),
                    )
        cir.plot()


def _prepare_initial_state(self: PQCombNet, is_choi_mode: bool) -> torch.Tensor:
    r"""
    Prepare the initial state for the quantum circuit.

    Args:
        is_choi_mode: A boolean flag indicating whether to prepare a Choi state.

    Returns:
        The initial state tensor.
    """
    ancilla_state = zero_state(
        num_systems=len(self.ancilla_dim_list), system_dim=self.ancilla_dim_list
    )
    bell_states = bell_state(
        num_systems=2 * (self.num_V if is_choi_mode else 1),
        system_dim=self.slot_dim,
    )
    return to_state(
        utils.linalg._nkron(ancilla_state.ket, bell_states.ket),
        ancilla_state.system_dim + bell_states.system_dim,
    )


def _construct_circuit(
    self: PQCombNet,
    system_dim_loss: int,
    unitary_set: torch.Tensor,
    V_list_applied_index: list,
) -> Circuit:
    r"""
    Construct the quantum circuit for the loss calculation.

    Args:
        system_dim_loss: The system dimension for the loss calculation.
        unitary_set: The set of unitary matrices to apply.
        V_list_applied_index: List of indices for applying V circuits.

    Returns:
        The constructed quantum circuit.
    """
    cir_loss = Circuit(system_dim=system_dim_loss)
    _apply_V_circuits(self, cir_loss, V_list_applied_index, unitary_set)
    return cir_loss


def _average_fidelity(
    self: PQCombNet,
    fid_type: str,
    projector: torch.Tensor = None,
) -> torch.Tensor:
    r"""
    Compute the average fidelity for a given set of unitaries and omega tensor.

    Args:
        fid_type: Type of fidelity calculation ('train' or 'test').
        projector: The projector to apply to the ancilla system of output state.
        max_batch_size: The maximum number of unitary matrices to be processed in parallel.

    Returns:
        The average fidelity as a real-valued tensor.
    """
    if fid_type == "test" and self.test_unitary_set.__len__() == 0:
        return torch.tensor(-1)
    # Set up the parameters
    is_comb_mode = self.train_mode == "comb"
    num_target_systems = 2 * self.num_V if is_comb_mode else 2
    system_dim_loss = self.ancilla_dim_list + [self.slot_dim] * num_target_systems
    V_list_applied_index = _get_V_list_applied_index(self, is_comb_mode)

    # Prepare the input state
    Psi_in = _prepare_initial_state(self, is_comb_mode)

    # Construct the circuit
    circuit = _construct_circuit(
        self,
        system_dim_loss,
        self.train_unitary_set if fid_type == "train" else self.test_unitary_set,
        V_list_applied_index,
    )

    # Compute the output density matrix
    psi_out = circuit(Psi_in).density_matrix @ (
        projector
        if projector is not None
        else torch.eye(self.ancilla_dim, dtype=Psi_in.dtype)
    ).kron(torch.eye(self.slot_dim**num_target_systems, dtype=Psi_in.dtype))

    psi_out_density_matrix = utils.linalg._partial_trace(
        psi_out,
        list(range(len(self.ancilla_dim_list))),
        system_dim_loss,
    )

    if is_comb_mode:
        return (
            utils.linalg._trace(
                psi_out_density_matrix
                @ (self.omega_train if fid_type == "train" else self.omega_test)
            )
            * self.slot_dim ** (self.num_V - 2)
        ).real
    else:
        return torch.mean(
            utils.linalg._trace(
                psi_out_density_matrix
                @ (self.omega_train if fid_type == "train" else self.omega_test)
            )
        ).real


def _save_results(
    self: PQCombNet,
    itr: int,
    fidelity: float,
    loss: float,
    base_lr: float,
    current_lr: float,
) -> None:
    data = {
        "slot_dim": self.slot_dim,
        "num_slots": self.num_slots,
        "ancilla_dim": self.ancilla_dim,
        "train_mode": self.train_mode,
        "base_lr": base_lr,
        "current_lr": current_lr,
        "num_test_unitary": len(self.train_unitary_set),
        "num_train_unitary": len(self.train_unitary_set),
        "seed": self.seed,
        "max_epochs": itr,
        "loss": loss,
        "fidelity": fidelity,
    }
    _save_results_to_csv(
        data=data,
        filename=f"{self.task_name}_train_log.csv",
        directory=self.data_directory_name,
    )


def _setup_parameters(
    self: PQCombNet,
    target_function: Callable[[torch.Tensor], torch.Tensor],
    num_slots: int,
    ancilla: Union[List[int], int] = 0,
    slot_dim: int = 2,
    train_mode: str = "comb",
    task_name: Optional[str] = None,
    is_ctrl_U: bool = False,
    seed: Optional[int] = None,
) -> None:
    r"""
    Combines the setup of basic parameters and training-specific parameters for quantum computation and training.

    Args:
        target_function: The function to apply to each unitary in the dataset
        num_slots: The number of unitaries to be queried
        ancilla: The ancilla dimension or dimension list
        slot_dim: The slot dimension for the unitaries to be queried
        train_mode: The training mode, which can be "process", "comb", or "swap"
        task_name: Optional name for the task, useful for data logging and storage
        is_ctrl_U: Flag to indicate if a controlled-U operation is used in the training process
        seed: Optional seed for random number generation, enhancing reproducibility
    """
    # Basic parameters setup
    self._target_function = target_function
    self._is_ctrl_U = is_ctrl_U
    self._slot_dim = slot_dim
    self._num_slots = num_slots
    self._ancilla_dim_list = (
        [2] * ancilla
        if isinstance(ancilla, int)
        else [dim for dim in ancilla if dim != 0]
    )
    self._num_V = num_slots + 1
    self._system_dim_list = self._ancilla_dim_list + [slot_dim]
    self._task_name = (
        task_name or f"pqcomb_{target_function.__name__}{'_ctrl' if is_ctrl_U else ''}"
    )

    # Training parameters setup
    self._seed = seed or np.random.randint(1e6)

    _validate_training_mode(self, train_mode)


def _setup_unitary_sets(
    self: PQCombNet,
    train_unitary_info: Union[int, torch.Tensor],
    test_unitary_info: Union[int, torch.Tensor],
) -> None:
    r"""
    Prepares the unitary sets for training and testing.

    Args:
        train_unitary_info: Information or data for training unitaries.
        test_unitary_info: Information or data for testing unitaries.
    """
    self._train_unitary_set = _generate_unitary_set(train_unitary_info, self.slot_dim)
    self._test_unitary_set = (
        _generate_unitary_set(test_unitary_info, self.slot_dim)
        if test_unitary_info != 0
        else torch.tensor([])
    )

    _calculate_omegas(self)


def _calculate_omegas(self: PQCombNet) -> None:
    r"""
    Calculate omega for train and test unitary sets.
    """
    try:
        self._omega_train = _get_omega(self, self.train_unitary_set)
        self._omega_test = _get_omega(self, self.test_unitary_set)
    except RuntimeError as e:
        if "not enough memory" not in str(e) or self.train_mode != "comb":
            raise e

        print(
            f"[{self.task_name} | {self.train_mode} | {self.seed}] "
            f"Out of memory error caught, switching train_mode from '{self.train_mode}' to ",
            end="",
        )
        self.train_mode = "process"
        print(f"'{self.train_mode}'...")
        self._omega_train = _get_omega(self, self.train_unitary_set)
        self._omega_test = _get_omega(self, self.test_unitary_set)


def _validate_training_mode(self: PQCombNet, train_mode: str) -> None:
    r"""
    Validates the training mode against supported modes.

    Args:
        train_mode: Training mode to validate.
    """
    train_mode_list = ["comb", "process", "swap"]
    if (train_mode := train_mode.lower()) not in train_mode_list:
        raise ValueError(
            f"Invalid train_mode: {train_mode}, must be one of {train_mode_list}"
        )
    if train_mode == "comb" and self.is_ctrl_U:
        raise ValueError("Controlled-U operation is not supported in 'comb' mode.")
    self._train_mode = train_mode


def _initialize_training_environment(self: PQCombNet) -> None:
    r"""
    Sets up the data directories and initializes the training environment.

    Initializes the omega tensors and the list of variable quantum circuits for the simulation.
    """
    self._data_directory_name = f"{self.task_name}_data"
    self._V_circuit_list = _create_V_circuit_list(self)


def _get_omega(self: PQCombNet, unitary_set: torch.Tensor) -> torch.Tensor:
    r"""
    Compute the omega tensor for a given task.

    Args:
        unitary_set: The set of unitary matrices.

    Returns:
        torch.Tensor: The omega tensor.
    """
    if unitary_set.__len__() == 0:
        return torch.tensor(-1)
    if self.train_mode == "comb":
        return _compute_omega_choi(
            self.target_function, unitary_set, self.slot_dim, self.num_slots
        )
    else:
        return _compute_omega_process(self.target_function, unitary_set, self.slot_dim)


def _create_V_circuit_list(self: PQCombNet) -> ModuleList:
    r"""
    Create a list of V circuits.

    Returns:
        ModuleList: The list of V circuits.
    """
    V_circuit_list = ModuleList()

    for _ in range(self.num_V):
        V = Circuit(system_dim=self.system_dim_list)
        V.universal_qudits(system_idx=list(range(len(self.system_dim_list))))
        V_circuit_list.append(V)

    return V_circuit_list


def _get_V_list_applied_index(
    self: PQCombNet, is_comb_mode: bool
) -> Union[List[List[int]], List[int]]:
    r"""
    Returns the list of indices where V circuits are applied, depending on the training mode.

    Args:
        is_choi_mode: Indicates whether the 'choi' mode is used for determining the index list.

    Returns:
        Union[List[List[int]], List[int]]: The list of indices where V circuits are applied.
    """
    if is_comb_mode:
        return [
            list(range(len(self.ancilla_dim_list)))
            + [len(self.ancilla_dim_list) + 2 * j + 1]
            for j in range(self.num_V)
        ]
    else:
        return list(range(len(self.system_dim_list)))


def _apply_V_circuits(
    self: PQCombNet,
    cir_loss: Circuit,
    V_list_applied_index: List[Union[List[int], int]],
    unitary_set: torch.Tensor,
) -> None:
    r"""
    Applies the V circuits to the circuit loss object.

    Args:
        cir_loss: The circuit to which the V circuits are applied.
        V_list_applied_index: A list or a list of lists of indices where V circuits should be applied.
        unitary_set: The set of unitary matrices used in the control operations.
    """
    for index, V_circuit in enumerate(self._V_circuit_list):
        cir_loss.oracle(
            V_circuit.matrix,
            (
                V_list_applied_index[index]
                if self.train_mode == "comb"
                else V_list_applied_index
            ),
            latex_name=f"\\mathcal{{V}}_{{{index}}}",
        )
        if self.train_mode != "comb" and index < self.num_slots:
            _apply_controlled_U(self, cir_loss, unitary_set, index)


def _apply_controlled_U(
    self: PQCombNet, cir_loss: Circuit, unitary_set: torch.Tensor, index: int
) -> None:
    r"""
    Applies the controlled U or U† depending on the configuration.

    Args:
        cir_loss: The circuit to which the controlled operations are applied.
        unitary_set: The set of unitary matrices.
        index: The index of the current operation in the sequence.
    """
    if self.is_ctrl_U:
        cir_loss.control_oracle(
            unitary_set if index % 2 == 0 else utils.linalg._dagger(unitary_set),
            [len(self.ancilla_dim_list) - 1, len(self.ancilla_dim_list)],
            latex_name=(r"U" if index % 2 == 0 else r"U^{\dagger}"),
        )
    else:
        cir_loss.oracle(
            unitary_set,
            len(self.ancilla_dim_list),
            latex_name=(r"U"),
        )


def _log_progress(
    self: PQCombNet,
    itr: int,
    loss: torch.Tensor,
    time_list: list,
    base_lr: float,
    current_lr: float,
    max_epochs: int,
    projector: torch.Tensor = None,
    is_save_data: bool = False,
    is_auto_stop: bool = True,
) -> bool:
    r"""
    Logs the training progress at specified intervals and saves the results.
    It provides insights into the current state of training, including loss, fidelity, and learning rate.

    This function checks if the current iteration is a multiple of 100 or the last iteration.
    If so, it calculates the average fidelity, constructs a log message with relevant metrics,
    and prints it. The function also saves the results and may determine if training should stop early.

    Args:
        itr: The current iteration number.
        loss: The current loss value.
        time_list: A list of time taken for each iteration.
        base_lr: The initial learning rate.
        current_lr: The current learning rate.
        max_epochs: The total number of iterations.
        projector: The projector to apply to the ancilla system of the output state.
        is_save_data: A flag to indicate whether to save the training data.
        is_auto_stop: A flag to indicate whether to stop training early if the learning rate is too low.

    Returns:
        Returns True if training should be stopped early, otherwise None.
    """
    if (
        itr % 100 == 0
        or itr == max_epochs - 1
        or (current_lr < 1e-3 * base_lr and is_auto_stop)
    ):
        fidelity = _average_fidelity(self, "test", projector).item()

        print(
            (
                f"[{self.task_name} | {self.train_mode} | {self.seed} | \033[90m{itr}\t{np.mean(time_list):.4f}s\033[0m] "
                f"slot_dim: {self.slot_dim}, slots: {self.num_slots}, "
                + (
                    f"ancilla_dim: {self.ancilla_dim}"
                    if all(dim == 2 for dim in self.ancilla_dim_list)
                    or not self.ancilla_dim_list
                    else f"aux_dim: {self.ancilla_dim}"
                )
                + f", \033[93mLR: {current_lr:.2e}\033[0m"
                + f", \033[91mLoss: {loss.item():.8f}\033[0m"
                + (f", \033[92mFid: {fidelity:.8f}\033[0m" if fidelity >= 0 else "")
            ),
        )

        time_list.clear()
        # Save results and possibly stop training
        if is_save_data:
            _save_results(self, itr, fidelity, loss.item(), base_lr, current_lr)

        # Stop training if auto-stop conditions are met
    return current_lr < 1e-3 * base_lr and is_auto_stop


def _compute_omega_choi(
    target_function: Callable[[torch.Tensor], torch.Tensor],
    unitary_set: torch.Tensor,
    slot_dim: int,
    num_slots: int,
) -> torch.Tensor:
    r"""
    Compute the omega tensor using the 'choi' mode.

    Args:
        target_function: The function to apply to each tensor in the dataset.
        unitary_set: The set of unitary matrices.
        slot_dim: The slot dimension for the unitaries to be queried.
        num_slots: The parameter 'num_slots' used in computation.

    Returns:
        torch.Tensor: The computed omega tensor.
    """
    omega = 0
    for u in unitary_set:
        u_transformed_choi = channel_repr_convert(
            target_function(u), source="kraus", target="choi"
        )
        u_conj_choi = channel_repr_convert(u.conj(), source="kraus", target="choi")
        omega += utils.linalg._nkron(
            *([u_transformed_choi] + [u_conj_choi] * num_slots)
        )
    omega /= len(unitary_set)
    perm_list = [0, 2 * num_slots + 1] + list(range(1, 2 * num_slots + 1))
    return utils.linalg._permute_systems(
        omega, perm_list, [slot_dim] * 2 * (num_slots + 1)
    )


def _generate_unitary_set(
    unitary_info: Union[int, torch.Tensor], slot_dim: int
) -> torch.Tensor:
    r"""
    Generates a set of unitary matrices based on provided info.

    Args:
        unitary_info: Details to generate or directly provide the unitaries.
        slot_dim: slot_dim for the unitary matrices.
    """
    return (
        random_unitary(num_systems=1, size=unitary_info, system_dim=slot_dim)
        if isinstance(unitary_info, int)
        else unitary_info
    )


def _compute_omega_process(
    target_function: Callable[[torch.Tensor], torch.Tensor],
    unitary_set: torch.Tensor,
    slot_dim: int,
) -> torch.Tensor:
    r"""
    Compute the omega tensor for the 'process' or 'swap' mode.

    Args:
        target_function: The function to apply to each unitary in the dataset.
        unitary_set: The set of unitary.
        slot_dim: The slot dimension for the unitaries to be queried.

    Returns:
        torch.Tensor: The computed omega tensor.
    """
    target_unitary = target_function(unitary_set)
    return (
        bell_state(num_systems=2, system_dim=slot_dim)
        .evolve(target_unitary.kron(torch.eye(slot_dim)))
        .density_matrix
    )


def _print_progress_bar(
    iteration: int,
    total: int,
    prefix: str = "",
    suffix: str = "",
    decimals: int = 1,
    length: int = 50,
    fill: str = "█",
) -> None:
    r"""
    Call in a loop to create a terminal progress bar.

    Args:
        iteration: Current iteration.
        total: Total iterations.
        prefix: Prefix string.
        suffix: Suffix string.
        decimals: Positive number of decimals in percent complete.
        length: Character length of the bar.
        fill: Bar fill character.
    """
    percent = f"{100 * (iteration / float(total)):.{decimals}f}"
    filled_length = int(length * iteration // total)
    bar = fill * filled_length + "-" * (length - filled_length)
    print(f"\r{prefix} |{bar}| {percent}% {suffix}", end="\r")
    if iteration == total:
        print()


def _save_results_to_csv(data: Dict[str, Any], filename: str, directory: str) -> None:
    r"""
    Save the results to a CSV file.

    Args:
        data: A dictionary containing the data to be saved.
        filename: The name of the CSV file.
        directory: The directory where the CSV file will be saved.
    """
    filepath = os.path.join(directory, filename)

    # Ensure the directory exists
    os.makedirs(directory, exist_ok=True)

    # Define the column names
    fieldnames = list(data.keys())

    # Write to the CSV file
    file_exists = os.path.isfile(filepath)

    with open(filepath, mode="a", newline="") as file:
        writer = csv.DictWriter(file, fieldnames=fieldnames)

        if not file_exists:
            writer.writeheader()

        writer.writerow(data)


def _extract_highest_fidelity(data_directory, filename):
    r"""
    Extract the highest fidelity for each combination of num_slots and ancilla_dim and generate a table.

    Args:
        data_directory: The directory where the CSV file is saved.
        filename: The name of the CSV file.
    """
    try:
        import pandas as pd
    except ImportError as e:
        raise ImportError(
            "The pandas library is required to run this function. Please install it using 'pip install pandas'."
        ) from e

    # Construct the full file path
    filepath = os.path.join(data_directory, filename)

    # Read the CSV file into a DataFrame
    df = pd.read_csv(filepath)

    # Find unique slot_dim values
    slot_dim_values = df["slot_dim"].unique()

    # Create a directory to save the result tables
    result_dir = os.path.join(data_directory, "fidelity_tables")
    os.makedirs(result_dir, exist_ok=True)

    # Iterate over each unique slot_dim value
    for slot_dim in slot_dim_values:
        # Filter the DataFrame for the current slot_dim
        df_filtered = df[df["slot_dim"] == slot_dim]

        # Pivot the table to have num_slots as rows and ancilla_dim as columns, with max fidelity as values
        pivot_table = df_filtered.pivot_table(
            index="num_slots",
            columns="ancilla_dim",
            values="fidelity",
            aggfunc="max",
        )

        # Save the pivot table to a CSV file
        output_filename = f"fidelity_table_slot_dim_{slot_dim}.csv"
        output_filepath = os.path.join(result_dir, output_filename)
        pivot_table.to_csv(output_filepath)
        print(f"Saved table for slot_dim = {slot_dim} to {output_filepath}")


def _qudit_swap_matrix(dim: int):
    """
    Returns the SWAP gate for a qudit of dimension dim.

    Args:
        dim (int): The dimension of the qudit.
    """
    qudit_swap_matrix = torch.zeros(dim**2, dim**2)
    for x, y in itertools.product(range(dim), range(dim)):
        qudit_swap_matrix[y * dim + x, x * dim + y] = 1
    return qudit_swap_matrix


def _train_swap_circuit(
    self: PQCombNet,
    depth: int = 1,
    base_lr: float = 0.1,
    max_epochs: int = 1000,
    loss_threshold: float = 1e-3,
) -> Tuple[Circuit, float]:
    r"""
    Train the SWAP gate circuit to pass the last unitary in the comb.

    Args:
        depth: The depth of the SWAP gate circuit.
        base_lr: The base learning rate for the optimizer.
        max_epochs: The maximum number of training epochs.
        loss_threshold: The loss threshold for stopping the training.

    Returns:
        Tuple[Circuit, float]: The trained SWAP gate circuit and the final loss value.
    """
    # check if the slot_dim is greater than or equal to the ancilla_dim
    assert self.slot_dim <= self.ancilla_dim, (
        f"slot_dim must be greater than or equal to ancilla_dim, "
        f"but got slot_dim={self.slot_dim} and ancilla_dim={self.ancilla_dim}"
    )

    assert self.slot_dim == self.ancilla_dim_list[-1], (
        f"slot_dim must be equal to the last element of ancilla_dim_list to construct the SWAP gate, "
        f"but got slot_dim={self.slot_dim} and ancilla_dim_list={self.ancilla_dim_list}"
    )
    actual_swap_cir = Circuit(system_dim=self.system_dim_list)
    for _ in range(depth):
        actual_swap_cir.universal_qudits(list(range(actual_swap_cir.num_systems)))
    ideal_swap_cir = Circuit(system_dim=self.system_dim_list)
    ideal_swap_cir.oracle(
        _qudit_swap_matrix(self.slot_dim),
        [len(self.system_dim_list) - 2, len(self.system_dim_list) - 1],
    )

    ideal_unitary_matrix = ideal_swap_cir.matrix
    ideal_choi_ket = (
        ideal_unitary_matrix.kron(torch.eye(self.system_dim))
        @ bell_state(2, self.system_dim).ket
    )

    # define the loss function as 1 minus the fidelity between the Choi matrices of the actual and desired circuits
    def swap_loss_func():
        actual_choi_ket = (
            actual_swap_cir.matrix.kron(torch.eye(self.system_dim))
            @ bell_state(2, self.system_dim).ket
        )
        # calculate the fidelity between the actual and desired density matrices
        return 1 - abs(utils.linalg._dagger(actual_choi_ket) @ ideal_choi_ket) ** 2

    # initialize the optimizer and scheduler
    opt = torch.optim.Adam(actual_swap_cir.parameters(), lr=base_lr)
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(opt, "min")

    # start the training process
    for itr in range(max_epochs):
        _print_progress_bar(itr + 1, max_epochs, prefix="Progress")
        # calculate the loss
        loss = swap_loss_func()
        # obtain the updated learning rate
        lr = scheduler.optimizer.param_groups[0]["lr"]
        if loss.item() < loss_threshold or lr < base_lr * 1e-4:
            print(
                f"[pqc_search_swap | \033[90m{itr}\033[0m | {depth}] ",
                f"Stop training SWAP with lr={lr:.2e}, loss={loss.item():.8f}.",
            )
            break

        # zero the gradients
        opt.zero_grad()
        # backpropagation
        loss.backward()
        # update the parameters
        opt.step()
        # adjust the learning rate according to the loss value
        scheduler.step(loss)

        # print training information every 40 iterations or on the last iteration
        if itr % 100 == 0 or itr == max_epochs - 1:
            print(
                f"[pqc_search_swap | \033[90m{itr}\033[0m | {depth}] "
                f"slot_dim: {self.slot_dim}, "
                + (
                    f"ancilla_dim: {self.ancilla_dim}"
                    if all(dim == 2 for dim in self.ancilla_dim_list)
                    or not self.ancilla_dim_list
                    else f"aux_dim: {self.ancilla_dim}"
                )
                + f", \033[93mLR: {lr:.2e}\033[0m, "
                f"\033[91mLoss: {loss.item():.8f}\033[0m"
            )

    if loss.item() < loss_threshold:
        print(f"Applying SWAP gate with lr={lr:.2e}, loss={loss.item():.8f}.")
        return actual_swap_cir, loss.item()
    else:
        print("Retrain SWAP gate...")
        return _train_swap_circuit(self, depth=depth + 1)


def _swap_train(
    self: PQCombNet,
    projector: Optional[torch.Tensor],
    base_lr: float,
    max_epochs: int,
    is_save_data: bool,
    is_auto_stop: bool,
) -> None:
    """
    Trains the PQCombNet model using the SWAP gate approach.

    This method incrementally adds parameterized SWAP gates to the model and trains them to enhance the model's capacity. The training process involves:
    - Initializing and training a SWAP gate.
    - Sequentially adding SWAP gates to the model until the desired number of slots is achieved.
    - Training only the newly added SWAP gates.
    - Training all gates in the model for fine-tuning.

    Optionally, it records training data such as fidelity metrics and timing information.

    Args:
        projector: The projector tensor used for fidelity calculations. If `None`, a default projector is used.
        base_lr: The base learning rate for the optimizer.
        max_epochs: The maximum number of training epochs for each training phase.
        is_save_data: If `True`, saves the training data and results to the specified directory.
        is_auto_stop: If `True`, enables early stopping based on convergence criteria.

    Returns:
        None.
    """
    n_s = self.num_slots
    print(
        f"Training SWAP gate with slot_dim={self.slot_dim} and ancilla_dim={self.ancilla_dim}"
    )
    start_time_overall = time.time()
    param_swap_gate, swap_loss = _train_swap_circuit(self)
    time_train_swap = time.time() - start_time_overall
    self.V_circuit_list = torch.nn.ModuleList(
        [Circuit(system_dim=self.system_dim_list)]
    )
    while self.num_slots < n_s:
        self.num_slots = self.V_circuit_list.__len__()
        for V_circuit in self.V_circuit_list:
            V_circuit.requires_grad_(False)
        self.V_circuit_list[-1].extend(deepcopy(param_swap_gate))
        self.V_circuit_list.append(deepcopy(param_swap_gate))
        if is_save_data:
            # Insert SWAP and calculate initial fidelity
            fidelity_train_swap_start = _average_fidelity(
                self, "train", projector
            ).item()
            fidelity_test_swap_start = _average_fidelity(self, "test", projector).item()

        # Train only SWAP gate
        print("Training only on parameterized SWAP gates.")
        start_time_swap_only = time.time()
        _train(
            self,
            projector,
            base_lr,
            max_epochs,
            is_save_data=False,
            is_auto_stop=is_auto_stop,
        )
        time_swap_only = time.time() - start_time_swap_only

        if is_save_data:
            # Calculate fidelity after SWAP-only training
            fidelity_train_swap_only = _average_fidelity(
                self, "train", projector
            ).item()
            fidelity_test_swap_only = _average_fidelity(self, "test", projector).item()

        # Allow all gates to be trained
        for V_circuit in self.V_circuit_list:
            V_circuit.requires_grad_(True)

        # Train all gates
        print("Training on all gates.")
        start_time_all = time.time()
        _train(
            self,
            projector,
            base_lr,
            max_epochs,
            is_save_data=is_save_data,
            is_auto_stop=is_auto_stop,
        )

        if is_save_data:
            time_all = time.time() - start_time_all

            # Calculate fidelity after training all gates
            fidelity_train_all = _average_fidelity(self, "train", projector).item()
            fidelity_test_all = _average_fidelity(self, "test", projector).item()

            if not os.path.exists(self.data_directory_name):
                os.makedirs(self.data_directory_name)

            csv_file_name = os.path.join(
                self.data_directory_name,
                f"swap_fidelity_time_dataset={self.train_unitary_set.__len__()}_d={self.slot_dim}.csv",
            )
            if not os.path.exists(csv_file_name):
                with open(csv_file_name, mode="w", newline="") as file:
                    writer = csv.writer(file)
                    writer.writerow(
                        [
                            "seed",
                            "num_slots",
                            "ancilla_dim",
                            "train_mode",
                            "fidelity_train_swap_start",
                            "fidelity_test_swap_start",
                            "time_swap_only",
                            "fidelity_train_swap_only",
                            "fidelity_test_swap_only",
                            "time_all",
                            "fidelity_train_all",
                            "fidelity_test_all",
                            "time_train_swap",
                            "time_overall",
                            "swap_loss",
                        ]
                    )

            swap_start_only_all_fidelity_time_item = (
                self.seed,
                self.num_slots,
                self.ancilla_dim,
                self.train_mode,
                fidelity_train_swap_start,
                fidelity_test_swap_start,
                time_swap_only,
                fidelity_train_swap_only,
                fidelity_test_swap_only,
                time_all,
                fidelity_train_all,
                fidelity_test_all,
                time_train_swap,
                time.time() - start_time_overall,
                swap_loss,
            )

            with open(csv_file_name, mode="a", newline="") as file:
                writer = csv.writer(file)
                writer.writerow(swap_start_only_all_fidelity_time_item)

    if is_save_data:
        # Extract the highest fidelity for each combination of num_slots and ancilla_dim
        try:
            import pandas as pd
        except ImportError as e:
            raise ImportError(
                "The pandas library is required to run this function. Please install it using 'pip install pandas'."
            ) from e

        # Read the CSV file into a DataFrame
        df = pd.read_csv(csv_file_name).round(4)

        # Pivot the table to have num_slots as rows and ancilla_dim as columns, with max fidelity_test_all as values
        pivot_table = df.pivot_table(
            index="num_slots",
            columns="ancilla_dim",
            values="fidelity_test_all",
            aggfunc="max",
        )

        # Save the pivot table to a new CSV file
        max_fidelity_filename = os.path.join(
            self.data_directory_name,
            f"max_fidelity_dataset={self.train_unitary_set.__len__()}_d={self.slot_dim}.csv",
        )
        pivot_table.to_csv(max_fidelity_filename)
        print(f"Saved max fidelity table to {max_fidelity_filename}")


def _train(
    self: PQCombNet,
    projector: Optional[torch.Tensor],
    base_lr: float,
    max_epochs: int,
    is_save_data: bool,
    is_auto_stop: bool,
) -> None:
    r"""
    Train the model using the provided parameters.

    Args:
        projector: The projector tensor used for fidelity calculations. If `None`, a default projector is used.
        base_lr: The base learning rate for the optimizer.
        max_epochs: The maximum number of training epochs.
        is_save_data: If `True`, saves the training data and results to the specified directory.
        is_auto_stop: If `True`, enables early stopping based on convergence criteria.

    Returns:
        None.
    """
    opt = torch.optim.Adam(self._V_circuit_list.parameters(), lr=base_lr)
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(opt, "min")

    time_list = []
    for itr in range(max_epochs):
        _print_progress_bar(itr + 1, max_epochs, prefix="Progress")

        start_time = time.time()
        loss = 1 - _average_fidelity(self, "train", projector)
        time_list.append(time.time() - start_time)

        if _log_progress(
            self,
            itr,
            loss,
            time_list,
            base_lr,
            scheduler.get_last_lr()[0],
            max_epochs,
            projector,
            is_save_data,
            is_auto_stop,
        ):
            break

        opt.zero_grad()
        loss.backward()
        opt.step()
        scheduler.step(loss)

    fidelity = _average_fidelity(self, "test", projector).item()

    if is_save_data:
        V_circuit_lists_dir = os.path.join(self.data_directory_name, "V_circuit_lists")
        os.makedirs(V_circuit_lists_dir, exist_ok=True)
        save_path = os.path.join(
            V_circuit_lists_dir,
            f"V_circuit_list_{self.train_mode}_sd{self.slot_dim}_ns{self.num_slots}_na{self.ancilla_dim}_itr{itr}"
            + (f"_fid{fidelity:.5f}.pt" if fidelity >= 0 else ".pt"),
        )
        torch.save(self.V_circuit_list, save_path)
    print(
        f"[{self.task_name} | {self.train_mode} | {self.seed}] Finished training"
        + (f" with Fidelity: {fidelity:.8f}" if fidelity >= 0 else "")
    )
